<?php

namespace Drupal\foldershare\Entity\FolderShareTraits;

use Drupal\Core\Database\Database;

use Drupal\Core\Entity\EntityStorageException;
use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\FolderShareInterface;
use Drupal\foldershare\ManageFileSystem;
use Drupal\foldershare\Settings;

/**
 * Recycle (Trash) FolderShare entities.
 *
 * This trait includes methods to recycle FolderShare entities from
 * the deleted FolderShare entities.
 *
 * <B>Internal trait</B>
 * This trait is internal to the FolderShare module and used to define
 * features of the FolderShare entity class. It is a mechanism to group
 * functionality to improve code management.
 *
 * @ingroup foldershare
 */
trait OperationRecycleTrait {

  /*---------------------------------------------------------------------
   *
   * Recycle.
   *
   *---------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function createRecycle() {

    //
    // it will store the deleted entities into the trash folder.
    //
    return $this->recycle();
  }

  /**
   * {@inheritdoc}
   */
  public function recycle() {
    if ($this->isRootItem() === TRUE) {
      // Move files and folders in the user's root list into the trash folder
      return $this->moveToRoot('', TRUE);
    }
    else {
      // Move files and folders in the user's root list into the trash folder
      $parent = $this->getParentFolder();
      if ($parent === NULL) {
        // There is no parent but it isn't a root item? Corrupted.
        return NULL;
      }

      return $this->moveToFolder($parent, '', TRUE);
    }
  }

  /**
   * Move multiple items in to the same parents or the user's root list
   * into the trash folder
   * @param array $ids
   * @return void
   * @throws EntityStorageException
   */
  public static function recycleMultiple(array $ids) {
    if (empty($ids) === TRUE) {
      // Nothing to move.
      return;
    }

    $recycleFolderName = Settings::getRecycleFolderName();
    $connection = Database::getConnection();

    //
    // Query the trash folder
    //
    $select = $connection->select(FolderShareInterface::BASE_TABLE, 'fs');
    $select->addField('fs', 'id', 'id');
    $select->condition('name', $recycleFolderName, '=');
    $select->condition('kind', 'folder', '=');
    $select->condition('uid', \Drupal::currentUser()->id(), '=');
    $select->isNull('fs.rootid');
    $select->isNull('fs.parentid');
    $results = $select->execute()->fetchAll();

    // The trash folder does not exist.
    // create trash folder
    if (count($results) === 0) {
      $recycleFolder = FolderShare::createRootFolder($recycleFolderName, false);
      $recycleFolder->setSystemHidden(TRUE);
      $recycleFolder->save();
    } else {
      $trashId = $results[0]->id;
      $recycleFolder = self::load($trashId);
      $trashFolderName = \Drupal::currentUser()->getAccountName() . '/' . $recycleFolderName;

      if (!ManageFileSystem::fileExists($trashFolderName)) {
        // Delete all items in the database
        FolderShare::deleteMultiple([$trashId], FALSE);
        // ManageFileSystem::createFolder($trashFolderName);
        $recycleFolder = FolderShare::createRootFolder($recycleFolderName, false);
        $recycleFolder->setSystemHidden(TRUE);
        $recycleFolder->save();
      }
    }

    self::moveToFolderMultiple($ids, $recycleFolder, TRUE);
    unset($recycleFolder);
  }

}
