<?php

namespace Drupal\foldershare\Form;

use Drupal\Component\Utility\Html;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Url;
use Drupal\foldershare\Settings;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\ConflictHttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Form\FormStateInterface;
use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\FolderShareInterface;

/**
 * Creates a form for the folder browser's ancestor menu.
 *
 * @ingroup foldershare
 */
class UIAncestorMenuField extends FormBase {

  /*--------------------------------------------------------------------
   *
   * Construction.
   *
   *--------------------------------------------------------------------*/
  const EMPTY_ITEM_ID = (-1);
  /**
   * The module handler, set at construction time.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;


  public function __construct(
    ModuleHandlerInterface $moduleHandler,
    FormBuilderInterface $formBuilder) {

    $this->moduleHandler = $moduleHandler;
    $this->formBuilder = $formBuilder;

  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('module_handler'),
      $container->get('form_builder')
    );
  }

  /*--------------------------------------------------------------------
   *
   * Form setup.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return mb_strtolower(str_replace('\\', '_', get_class($this)));
  }

  public function buildForm(array $form, FormStateInterface $formState)
  {

    //
    // Validate destination.
    // ---------------------
    // Get the destination ID from the form state, or use a default.

    $args = $formState->getBuildInfo()['args'];
    if (empty($args) === TRUE) {
      $destinationId = FolderShareInterface::USER_ROOT_LIST;
    }
    else {
      $destinationId = intval($args[0]);
    }

    // If there is prior form state, and it includes a destination ID,
    // then get it. The ID might still be NULL, or it might be a word
    // naming a root list.
    if ($formState !== NULL) {
      $userInput = $formState->getUserInput();
      if (isset($userInput['destinationid']) === TRUE) {
        // The destination ID coming back from the UI could be an integer
        // ID, or one of several words like "home", "shared", "public",
        // and "all", to indicate specific root lists.
        $destinationId = $userInput['destinationid'];
      }
    }

    //
    // Load entity.
    // ------------
    // Load the current entity, if any.
    $currentItem = NULL;
    if ($destinationId >= 0) {
      $currentItem = FolderShare::load($destinationId);
      if ($currentItem === NULL) {
        $destinationId = FolderShareInterface::USER_ROOT_LIST;
      }
    }

    $uiClass         = 'foldershare-ancestormenu';
    $menuButtonClass = 'foldershare-ancestormenu-menu-button';
    $menuClass       = 'foldershare-ancestormenu-menu';

    $menuButtonUrl   = '/' . \Drupal::service('extension.path.resolver')->getPath('module', 'foldershare') .
      '/images/menu-ancestor.svg';
    $menuButtonAlt   = t('Location menu');
    $menuButtonImage = "<img alt=\"$menuButtonAlt\" src=\"$menuButtonUrl\">";

    //
    // Add ancestor list markup.
    // -------------------------
    // Build a list of ancestor folders for the ancestor menu. For each one,
    // include the ancestor's URL as an attribute. Javascript will use this
    // to load the appropriate page. The URL is not included in an <a> for
    // the menu item so that menu items aren't styled as links.
    //
    // File classes are added so that themes can style the item with
    // folder icons.
    //
    // Drupal work-around: There are no render array #type's to make
    // hierarchical lists. Even if an abstract list type existed, it would
    // be routed through theme templates that could alter the structure,
    // strip classes, etc. Since we need a very specific structure with
    // specific classes and data attributes, the only practical way to
    // build this is with explicit HTML.
    $menuMarkup = "<ul class=\"hidden $menuClass\">";

    if ($currentItem !== NULL) {
      // The page is for entity. Get its ancestors.
      $folders = $currentItem->findAncestorFolders();

      // Reverse ancestors from root first to root last.
      $folders = array_reverse($folders);

      // Push the current entity onto the ancestor list so that it gets
      // included in the menu.
      array_unshift($folders, $currentItem);

      // Add ancestors to menu.
      foreach ($folders as $item) {
        if (str_starts_with($item->getPath(), '/'. Settings::getRecycleFolderName()) &&
          trim($item->getName()) === Settings::getRecycleFolderName()) {
          continue;
        }

        // Get the URL to the folder.
        $entity_uuid = $item->uuid();
        $rel = 'canonical';
        $entityTypeId = $item->getEntityTypeId();
        $url = rawurlencode(Url::fromRoute("entity.{$entityTypeId}.$rel",
          [$entityTypeId => $entity_uuid])
          ->setAbsolute()
          ->toString());

        // Get the name for the folder.
        $name = Html::escape($item->getName());

        // Add the HTML. Include file classes that mark this as a folder
        // or file.
        if ($item->isFolder() === TRUE) {
          $fileClasses = 'file file--folder file--mime-folder-directory';
        }
        else {
          $mimes = explode('/', $item->getMimeType());
          $fileClasses = 'file file--' . $mimes[0] .
            ' file--mime-' . $mimes[0] . '-' . $mimes[1];
        }

        if ($item->isSystemDisabled() === TRUE) {
          $attr = '';
        }
        else {
          $attr = 'data-foldershare-id="' . $item->id() . '"';
        }

        $menuMarkup .= "<li $attr data-foldershare-url=\"$url\"><div><span class=\"$fileClasses\"></span>$name</div></li>";
      }
    }
    $menuMarkup .= '</ul>';

    //
    // Create UI
    // ---------
    // Everything is hidden initially, and only exposed by Javascript, if
    // the browser supports Javascript.
    $form['#attributes']['class'][] = 'foldershare-folder-ancestor-menu-form';
    $form['ui-ancestor-menu'] = [
      '#attributes' => [
        'class'     => [
          'foldershare-ancestormenu',
        ],
      ],
      $uiClass => [
        '#type'              => 'container',
        '#weight'            => -90,
        '#attributes'        => [
          'class'            => [
            $uiClass,
            'hidden',
          ],
        ],

        // Add a hierarchical menu of ancestors. Javascript uses jQuery.ui
        // to build a menu from this and presents it from a menu button.
        // The menu was built and marked as hidden.
        //
        // Drupal work-around: The field is built using an inline template
        // to avoid Drupal's HTML cleaning that can remove classes and
        // attributes on the menu items, which we need to retain to provide
        // the URLs of ancestor folders. Those URLs are used by Javascript
        // to load the appropriate page when a menu item is selected.
        $menuClass           => [
          '#type'            => 'inline_template',
          '#template'        => '{{ menu|raw }}',
          '#context'         => [
            'menu'           => $menuMarkup,
          ],
        ],

        // Add a button to go up to an ancestor folder. Javascript binds
        // a behavior to the button to load an ancestor page. The button is
        // hidden initially and only shown if the browser supports Javascript.
        //
        // Drupal work-around: If we use the #type "button" or "image_button",
        // we'll get a button, but not necessarily using HTML's <button>.
        // This varies by theme. The core themes, based on "classy",
        // implement buttons with <input>. Some Bootstrap-based themes
        // implement buttons with <button>, while others do not.
        // Unfortunately, <input> does not support wrapping an <img> for
        // an icon, while <button> does (which makes it odd that core's
        // #type "image_button" still generates <input>, which will never
        // work).
        //
        // To guarantee that we get an <img> on the button, we have to
        // force use of <button>, and that requires using #type "html_tag".
        // Unfortunately, "html_tag" only allows a string as the value
        // within the tag, so we are forced to enter explicit HTML below.
        //
        // A downside of using "html_tag" is that the button does not get
        // processed by core's #type "button" code, a theme's button
        // template pre-processor, or a theme's button template. Since those
        // are responsible for adding button classes, we have to add them
        // here explicitly:
        //
        // - "button", "js-form-submit", and "form-submit" are added by
        //   core #type "button".
        //
        // - "button--image-button" is added by core #type "button" if the
        //   #button_type is "image_button".
        //
        // - "btn" is added by many Bootstrap-based themes.
        //
        // - "w3-button", "w3-border", and "w3-theme-border" are W3.CSS
        //   framework standards.
        //
        // While it is possible to create a custom button template for this
        // kind of button, core's themeing system always cascades from
        // simplest to most custom template, invoking ALL of them. Because
        // the core themes do not define twig blocks around <input>, it is
        // not possible for a custom theme late in the cascade to override
        // the <input> tag created earlier in the cascade. This is why
        // custom themes (like Bootstrap) do not extend the core themes.
        // But here we don't have that option - we have to work with any
        // theme, including core themes, and therefore we cannot use a
        // custom template to override button handling.
        $menuButtonClass     => [
          '#type'            => 'html_tag',
          '#tag'             => 'button',
          '#value'           => $menuButtonImage,
          '#name'            => $menuButtonClass,
          '#attributes'      => [
            'title'          => t('See the location of the current item.'),
            'class'          => [
              $menuButtonClass,
              'hidden',
              // Core classes.
              'button',
              'button--image-button',
              'js-form-submit',
              'form-submit',
              // Bootstrap classes.
              'btn',
              // W3.CSS classes.
              'w3-button',
              'w3-border',
              'w3-theme-border',
              // jQuery.UI. With some themes, these get lost when set by
              // jQuery using button(). Why?
              'ui-button',
              'ui-corner-all',
              'ui-widget',
            ],
          ],
        ],
      ],

      'hiddengroup'         => [
        '#type'             => 'container',
        '#attributes'       => [
          'class'           => [
            'hidden',
          ],
        ],

        // Include a hidden text field filled in by Javascript as the user
        // selects new destination folders by double-clicks or selecting from
        // the ancestor menu.
        'destinationid'     => [
          '#type'           => 'textfield',
          '#name'           => 'destinationid',
          '#default_value'  => $destinationId,
        ],

        // Include a hidden refresh button "clicked" by Javascript each time the
        // destination ID above is changed. Below we add AJAX callbacks
        // to trigger a refresh of this part of the form.
        'refresh'           => [
          '#type'           => 'button',
          '#name'           => 'refresh',
          '#value'          => 'Refresh',
        ],

        // Include a hidden text field filled in by Javascript if the user
        // selects, but does not double-click, a folder in the list. This
        // becomes the selected destination. If nothing is selected, then
        // the parent of the current view is the destination.
        'selectionid'       => [
          '#type'           => 'textfield',
          '#name'           => 'selectionid',
          '#default_value'  => self::EMPTY_ITEM_ID,
        ],
      ],

    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(
    array &$form,
    FormStateInterface $formState) {

    //
    // Validate trigger.
    // -----------------
    // Ignore triggers other than 'submit' and 'refresh'. They should not
    // occur anyway.
    $triggerName = $formState->getTriggeringElement()['#name'];
    if ($triggerName !== 'refresh' && $triggerName !== 'submit') {
      return;
    }

    //
    // Validate form input.
    // --------------------
    // Get the user's input to the form, if any, and the current parent.
    $userInput     = $formState->getUserInput();
    $destinationId = $userInput['destinationid'];

    if (empty($destinationId)) {
      $args = $formState->getBuildInfo()['args'];
      if (empty($args) === TRUE) {
        $destinationId = FolderShareInterface::USER_ROOT_LIST;
      } else {
        $destinationId = intval($args[0]);
      }
    }

    $selectionId   = $userInput['selectionid'];

    if ($selectionId === NULL) {
      $selectionId = self::EMPTY_ITEM_ID;
    }

    // If the selection ID is NULL or negative, then there is no selection.
    // Default to the currently shown folder, which is in the destination ID.
    if ($selectionId >= 0) {
      if ($triggerName === 'submit' && $selectionId >= 0) {
        $destinationId = $selectionId;
      }
      else {
        // $destinationId = $this->getParentId();
        // if ($destinationId < 0) {
          $destinationId = FolderShareInterface::USER_ROOT_LIST;
        // }
      }
    }

    // Convert special root list names to negative numbers that flag the
    // equivalent root lists. This lets us pass around and compare integers,
    // rather than consider string values as well.
    switch ($destinationId) {
      case FolderShareInterface::USER_ROOT_LIST:
      case 'home':
        $destinationId = FolderShareInterface::USER_ROOT_LIST;
        break;

      case FolderShareInterface::TRASH_ROOT_LIST:
      case 'trash':
        $destinationId = FolderShareInterface::TRASH_ROOT_LIST;
        break;

      case FolderShareInterface::SHARED_ROOT_LIST:
      case 'shared':
        $destinationId = FolderShareInterface::SHARED_ROOT_LIST;
        break;

      case FolderShareInterface::PUBLIC_ROOT_LIST:
      case 'public':
        // The public root list is not supported. Map it to the home folder.
        $destinationId = FolderShareInterface::USER_ROOT_LIST;
        break;

      case FolderShareInterface::ALL_ROOT_LIST:
      case 'all':
        $destinationId = FolderShareInterface::ALL_ROOT_LIST;
        break;

      default:
        if (is_numeric($destinationId) === FALSE) {
          // Developer-facing exception message.
          throw new NotFoundHttpException(
            "A top-level item with ID '$destinationId' could not be found.");
        }

        $destinationId = (int) $destinationId;
        if ($destinationId < 0) {
          $destinationId = FolderShareInterface::USER_ROOT_LIST;
        }
        else {
          $destination = FolderShare::load($destinationId);
          if ($destination === NULL ||
              $destination->isSystemHidden() === TRUE) {

            if ($destination->getName() != Settings::getRecycleFolderName()) {
              throw new NotFoundHttpException(
                FolderShare::getStandardHiddenMessage($destination->getName()));
            }
          }

          if ($destination->isSystemDisabled() === TRUE) {
            throw new ConflictHttpException(
              FolderShare::getStandardDisabledMessage(
                'accessed',
                $destination->getName()));
          }
        }
        break;
    }

    //
    // Dispatch based on trigger.
    // --------------------------
    // For a refresh, rebuild the form. For a submit, validate it.
    if ($triggerName === 'refresh') {
      $formState->setRebuild(TRUE);
      // return;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(
    array &$form,
    FormStateInterface $formState) {
  }

}
