<?php

namespace Drupal\foldershare\Form;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\foldershare\Settings;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\ConflictHttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Form\FormStateInterface;
use Drupal\foldershare\Constants;
use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\FolderShareInterface;

/**
 * Creates a form for the folder browser's folder and file list.
 *
 * @ingroup foldershare
 */
class UIViewsField extends FormBase {

  /*--------------------------------------------------------------------
   *
   * Construction.
   *
   *--------------------------------------------------------------------*/
  const EMPTY_ITEM_ID = (-1);
  /**
   * The module handler, set at construction time.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;


  public function __construct(
    ModuleHandlerInterface $moduleHandler,
    FormBuilderInterface $formBuilder) {

    $this->moduleHandler = $moduleHandler;
    $this->formBuilder = $formBuilder;

  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('module_handler'),
      $container->get('form_builder')
    );
  }

  /*--------------------------------------------------------------------
   *
   * Form setup.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return mb_strtolower(str_replace('\\', '_', get_class($this)));
  }

  public function buildForm(array $form, FormStateInterface $formState)
  {

    //
    // Validate destination.
    // ---------------------
    // Get the destination ID from the form state, or use a default.
    // $destinationId = NULL;

    $args = $formState->getBuildInfo()['args'];
    if (empty($args) === TRUE) {
      $viewName = Constants::VIEW_LISTS;
      $displayName = Constants::VIEW_DISPLAY_LIST_FOLDER;
      $destinationId = FolderShareInterface::USER_ROOT_LIST;
      $style = '';
    }
    else {
      $viewName = $args[0];
      $displayName = $args[1];
      $destinationId = intval($args[2]);
      $style = $args[3];
    }

    // If there is prior form state, and it includes a destination ID,
    // then get it. The ID might still be NULL, or it might be a word
    // naming a root list.
    if ($formState !== NULL) {
      $userInput = $formState->getUserInput();
      if (isset($userInput['destinationid']) === TRUE) {
        $destinationId = $userInput['destinationid'];

        // The destination ID coming back from the UI could be an integer
        // ID, or one of several words like "home", "shared", "public",
        // and "all", to indicate specific root lists.
      }
    }

    //
    // Create UI
    // ---------
    // Everything is hidden initially, and only exposed by Javascript, if
    // the browser supports Javascript.
    $form['#attributes']['class'][] = 'foldershare-folder-view-browser-form';
    $form['ui-foldershare-view-browser'] = [
      '#attached' => [
        'drupalSettings' => [
          'foldershare-view-page' => [
            'viewName'        => $viewName,
            'displayName'     => $displayName,
            'viewAjaxEnabled' => TRUE,
          ],
        ],
      ],

      'folder-view-browser' => [
        '#type'      => 'container',
        '#attributes' => [
          'class' => [
            'foldershare-sidebar-and-folder-browser',
          ],
        ],

        'hiddengroup'         => [
          '#type'             => 'container',
          '#attributes'       => [
            'class'           => [
              'hidden',
            ],
          ],

          // Include a hidden text field filled in by Javascript as the user
          // selects new destination folders by double-clicks or selecting from
          // the ancestor menu.
          'destinationid'     => [
            '#type'           => 'textfield',
            '#name'           => 'destinationid',
            '#default_value'  => $destinationId,
          ],

          // Include a hidden refresh button "clicked" by Javascript each time the
          // destination ID above is changed. Below we add AJAX callbacks
          // to trigger a refresh of this part of the form.
          'refresh'           => [
            '#type'           => 'button',
            '#name'           => 'refresh',
            '#value'          => 'Refresh',
          ],

          // Include a hidden text field filled in by Javascript if the user
          // selects, but does not double-click, a folder in the list. This
          // becomes the selected destination. If nothing is selected, then
          // the parent of the current view is the destination.
          'selectionid'       => [
            '#type'           => 'textfield',
            '#name'           => 'selectionid',
            '#default_value'  => self::EMPTY_ITEM_ID,
          ],
        ],

        // Add the view.
        'view'          => [
          '#type'       => 'view',
          '#embed'      => TRUE,
          '#name'       => $viewName,
          '#display_id' => $displayName,
          '#arguments'  => [$destinationId],
          '#attributes' => [
            'class'     => [
              'foldershare-folder-browser',
            ],
            'style'     => [
              $style,
            ],
          ],
        ],
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(
    array &$form,
    FormStateInterface $formState) {

    //
    // Validate trigger.
    // -----------------
    // Ignore triggers other than 'submit' and 'refresh'. They should not
    // occur anyway.
    $triggerName = $formState->getTriggeringElement()['#name'];
    if ($triggerName !== 'refresh' && $triggerName !== 'submit') {
      return;
    }

    //
    // Validate form input.
    // --------------------
    // Get the user's input to the form, if any, and the current parent.
    $userInput     = $formState->getUserInput();
    $destinationId = $userInput['destinationid'];

    if (empty($destinationId)) {
      $args = $formState->getBuildInfo()['args'];
      if (empty($args) === TRUE) {
        $destinationId = FolderShareInterface::USER_ROOT_LIST;
      }
      else {
        $destinationId = intval($args[2]);
      }
    }

    $selectionId   = $userInput['selectionid'];
    if ($selectionId === NULL) {
      $selectionId = self::EMPTY_ITEM_ID;
    }

    // If the selection ID is NULL or negative, then there is no selection.
    // Default to the currently shown folder, which is in the destination ID.
    if ($selectionId >= 0) {
      if ($triggerName === 'submit' && $selectionId >= 0) {
        $destinationId = $selectionId;
      }
      else {
        // $destinationId = $this->getParentId();
        // if ($destinationId < 0) {
          $destinationId = FolderShareInterface::USER_ROOT_LIST;
        // }
      }
    }

    if (is_numeric($destinationId) === FALSE) {
      // Developer-facing exception message.
      throw new NotFoundHttpException(
        "A top-level item with ID '$destinationId' could not be found.");
    }

    $destinationId = (int) $destinationId;
    if ($destinationId < 0) {
      $destinationId = FolderShareInterface::USER_ROOT_LIST;
    }
    else {
      $destination = FolderShare::load($destinationId);
      if ($destination === NULL ||
          $destination->isSystemHidden() === TRUE) {
        if ($destination->getName() != Settings::getRecycleFolderName()) {
          throw new NotFoundHttpException(
            FolderShare::getStandardHiddenMessage($destination->getName()));
        }
      }

      if ($destination->isSystemDisabled() === TRUE) {
        throw new ConflictHttpException(
          FolderShare::getStandardDisabledMessage(
            'accessed',
            $destination->getName()));
      }
    }

    //
    // Dispatch based on trigger.
    // --------------------------
    // For a refresh, rebuild the form. For a submit, validate it.
    if ($triggerName === 'refresh') {
      $formState->setRebuild(TRUE);
      // return;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(
    array &$form,
    FormStateInterface $formState) {
    // the empty action
  }

}
