<?php

namespace Drupal\foldershare\Form;

use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

use Drupal\field\Entity\FieldConfig;
use Drupal\foldershare\FolderShareInterface;
use Drupal\foldershare\ManageFileSystem;
use Drupal\foldershare\ManageS3FileSystem;
use Drupal\foldershare\Utilities\FileUtilities;
use Symfony\Component\DependencyInjection\ContainerInterface;

use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\Entity\FolderShareScheduledTask;
use Drupal\foldershare\ManageSearch;
use Drupal\foldershare\ManageUsageStatistics;
use Drupal\foldershare\Settings;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Provides a form for removing all files and folders before an uninstall.
 *
 * <B>Access control</B>
 * The route to this form should restrict access to those with administration
 * permission.
 *
 * @ingroup foldershare
 */
class UninstallFolderShareConfirm extends ConfirmFormBase {

  /*--------------------------------------------------------------------
   *
   * Fields - dependency injection.
   *
   *--------------------------------------------------------------------*/
  /**
   * The entity type manager, set at construction time.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The item list to be deleted.
   *
   * @var array
   */
  protected array $itemsToDelete;

  /*--------------------------------------------------------------------
   *
   * Construction.
   *
   *--------------------------------------------------------------------*/
  /**
   * Constructs a new page.
   *
   * @param \Drupal\Core\Entity\EntityTypeManager $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManager $entityTypeManager) {
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static($container->get('entity_type.manager'));
  }

  /*--------------------------------------------------------------------
   *
   * Form setup.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return mb_strtolower(str_replace('\\', '_', get_class($this)));
  }

  /**
   * Returns the title for the form.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The form title.
   */
  public function getTitle() {
    return $this->getQuestion();
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Are you sure you want to delete all files and folders?');
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    // The description is filled in during form build.
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText() {
    return $this->t('Delete all files and folders');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return Url::fromRoute('system.modules_uninstall');
  }

  /*--------------------------------------------------------------------
   *
   * Form build.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $formState = NULL) {
    $form = parent::buildForm($form, $formState);

    // Get the number of items to delete. This is an expensive query,
    // so we didn't do it in getDescription().
    $n = FolderShare::countNumberOfItems();

    // Print the warning notice for deleting the foldershare
    $noticeFlag = TRUE;
    // Now that we have the number of items to delete, create a better
    // description.
    if ($n === 0) {
      $noticeFlag = FALSE;
      $form['description']['#markup'] = $this->t(
        '<p>There are no files and folders to delete.</p>');

      // Hide the delete button if there is nothing to delete.
      $form['actions']['submit']['#access'] = FALSE;
    }
    else {
      $form['description']['#markup'] = $this->t(
        '<p>There are @n FolderShare module files and folders to delete.</p>',
      [
        '@n' => $n,
      ]);
    }

    $entityTypeManager = \Drupal::service('entity_type.manager');
    $entity_type_ids = ['node' => 'node_type'];
    foreach ($entity_type_ids as $field_type_id => $entity_type_id) {
      $contentTypes = $entityTypeManager->getStorage($entity_type_id)
        ->loadMultiple();
      foreach ($contentTypes as $contentType) {
        $type = $contentType->id();
        $fields = array_filter(
          \Drupal::service('entity_field.manager')
            ->getFieldDefinitions($field_type_id, $type),
          function ($fieldDefinition) {
            return $fieldDefinition instanceof \Drupal\field\FieldConfigInterface;
          }
        );

        // String to search in entity_reference_revisions
        $searchStr = 'foldershare';
        foreach ($fields as $key => $definition) {
          if ($definition->getType() == "foldershare_field") {
            $this->itemsToDelete[$type][] = $key;
          }
          if ($definition->getType() == "entity_reference_revisions" && str_contains($key, $searchStr)) {
            $this->itemsToDelete[$type][] = $key;
          }
        }
      }
    }

    $weight = 0;
    if(isset($this->itemsToDelete)) {
      $noticeFlag = TRUE;
      $form['title'] = [
        '#markup' => $this->t('<div><b>FOLDERSHARE FIELDS</b></div>'),
        '#weight' => $weight++,
      ];
      $form['break1'] = [
        '#markup' => '<hr>',
        '#weight' => $weight++,
      ];

      $body = "<ul>";

      foreach ($this->itemsToDelete as $type => $fieldsArray) {
        foreach ($fieldsArray as $fieldName) {
          $body .= "<li> $fieldName ($type) </li>";
        }
      }

      $body .= "</ul>";
      $form['body'] = [
        '#markup' => $this->t("<div>$body</div>"),
        '#weight' => $weight++,
      ];
    }

    if ($noticeFlag) {
      $form['notice'] = [
        '#markup' => $this->t("<p>Deleting them cannot be undone. Make a backup of your database if you want to be able to restore these items. This only affects files managed by the FolderShare module. Files used by other modules are not affected.</p>"),
        '#weight' => $weight++,
        ];
    }

    return $form;
  }

  /*--------------------------------------------------------------------
   *
   * Form validate.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $formState) {
    // Nothing to do.
  }

  /*--------------------------------------------------------------------
   *
   * Form submit.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function submitForm(array &$form, FormStateInterface $formState) {
    // Stop all background tasks first.
    // - Stop the task scheduler.
    // - Stop the search indexer.
    // - Stop the usage table updater.
    // - Disable activity logging.
    $taskEnabled = FolderShareScheduledTask::isTaskExecutionEnabled();
    FolderShareScheduledTask::setTaskExecutionEnabled(FALSE);

    $indexInterval = Settings::getSearchIndexInterval();
    Settings::setSearchIndexInterval('system');

    $usageInterval = Settings::getUsageUpdateInterval();
    Settings::setUsageUpdateInterval('manual');

    $loggingEnable = Settings::getActivityLogEnable();
    Settings::setActivityLogEnable(FALSE);

    // Delete all tasks.
    FolderShareScheduledTask::deleteAllTasks();

    // Delete all files and folders.
    FolderShare::deleteAll(FolderShareInterface::ANY_USER_ID,TRUE);

    // Clear the search index.
    ManageSearch::clearIndex();

    // Clear usage table.
    ManageUsageStatistics::clearUsage();

    // It is possible that the user will NOT continue the uninstall after
    // running this form to delete everything. In that case we need to leave
    // the module in a working state.
    // - Restore activity logging.
    // - Restore the usage table updater.
    // - Restore the search indexer.
    // - Restore the task scheduler.
    //
    // Let the module's preuninstall() hook finish cleanup by deleting
    // the directory tree containing local files, if the user continues
    // to uninstall the module.
    Settings::setUsageUpdateInterval($usageInterval);

    Settings::setSearchIndexInterval($indexInterval);

    Settings::setActivityLogEnable($loggingEnable);

    FolderShareScheduledTask::setTaskExecutionEnabled($taskEnabled);

    // Remove foldershare fields
    $this->removeFolderShareFields();

    // Delete all files and directories created by the module.
    $fileScheme = Settings::getFileScheme();
    $uri = $fileScheme . '://' . ManageFileSystem::FILE_DIRECTORY;
    if ($fileScheme === 's3') {
      ManageS3FileSystem::deleteS3Directory(ManageFileSystem::FILE_DIRECTORY);
    } else {
      FileUtilities::rrmdir($uri);
    }

    $formState->setRedirect('system.modules_uninstall');
  }

  /**
   * Remove the foldershare fields
   *
   * @return void
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function removeFolderShareFields() {
    $is_deleted = FALSE;
    $entity_type_ids = ['node'];
    if(isset($this->itemsToDelete)) {
      foreach ($this->itemsToDelete as $type => $fieldsArray) {
        foreach ($fieldsArray as $fieldName) {
          foreach ($entity_type_ids as $entity_id) {
            $is_deleted = TRUE;
            $field = FieldConfig::loadByName($entity_id, $type, $fieldName);
            $field?->delete();
          }
        }
      }

      if($is_deleted) {
        // Run cron job to clean up the fields
        $cron = \Drupal::service('cron');
        $cron->run();
      }

      $this->messenger()->addMessage($this->t('All FolderShare fields deleted'));
    }
    else {
      $this->messenger()->addMessage($this->t('Nothing to delete'));
    }
  }
}
