<?php

namespace Drupal\foldershare;

use Drupal\Component\Render\PlainTextOutput;
use Drupal\Core\Database\Database;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\Utilities\LimitUtilities;
use Drupal\views\Views;

/**
 * Manages content types for FolderShare entities.
 *
 * This class provides static methods to manage creating, updating, deleting, and retrieving
 * foldershare fields for the content types.
 * It also retrieves the configuration for content type about the foldershare entity.
 *
 * Supported operations include:
 * - Clearing the foldershare fields in content type.
 * - Deleting the foldershare fields in content type.
 * - Getting the foldershare fields in content type.
 * - Updating the foldershare fields in content type.
 *
 * This method also provides methods to query configuration settings for
 * foldershare in content type.
 *
 * <B>Access control</B>
 * This class's methods do not do access control. The caller should check
 * access as needed by their situation.
 *
 * @ingroup foldershare
 *
 * @see \Drupal\foldershare\Entity\FolderShare
 */
final class ManageContentType {

  /*--------------------------------------------------------------------
   *
   * Constants.
   *
   *-------------------------------------------------------------------*/
  /**
   * The name of the content type tracking database table.
   */
  const CONTENT_TYPE_TABLE = 'foldershare_content_type';

  /*--------------------------------------------------------------------
   *
   * Table definition.
   *
   *-------------------------------------------------------------------*/
  /**
   * Returns the database schema that defines the content type for the foldershare table.
   *
   * The table contains one record for content type. Record fields include:
   *
   * - The user ID.
   * - The entity bundle.
   * - The entity type.
   * - The entity ID.
   * - The foldershare ID.
   * - The foldershare UUID.
   * - The created date.
   * - The modified date.
   *
   * @return array
   *   Returns an array describing the above table.
   */
  public static function getSchema() {
    $schema[self::CONTENT_TYPE_TABLE] = [
      'description' => 'Stores file and folder usage information for users.',

      'fields' => [
        // The user ID for the user.
        'user_id' => [
          'type' => 'int',
          'unsigned' => TRUE,
          'size' => 'big',
          'not null' => TRUE,
          'default' => 0,
          'description' => 'The user ID.',
        ],

        // bundle.
        'bundle' => [
          'type' => 'varchar',
          'length' => 128,
          'not null' => TRUE,
          'default' => '',
          'description' => 'The bundle.',
        ],

        // entity type.
        'entity_type' => [
          'type' => 'varchar',
          'length' => 128,
          'not null' => TRUE,
          'default' => '',
          'description' => 'The entity type.',
        ],

        // entity id.
        'entity_id' => [
          'type' => 'int',
          'unsigned' => TRUE,
          'size' => 'big',
          'not null' => TRUE,
          'default' => 0,
          'description' => 'The entity ID.',
        ],

        // The foldershare ID.
        'fs_id' => [
          'type' => 'int',
          'unsigned' => TRUE,
          'size' => 'big',
          'not null' => TRUE,
          'default' => 0,
          'description' => 'The foldershare ID.',
        ],

        // The foldershare UUID.
        // entity type.
        'fs_uuid' => [
          'type' => 'varchar',
          'length' => 128,
          'not null' => TRUE,
          'default' => '',
          'description' => 'The foldershare UUID.',
        ],

        'created' => [
          'mysql_type' => 'timestamp',
          'pgsql_type' => 'timestamp',
          'sqlite_type' => 'datetime',
          'not null' => FALSE,
          'default' => NULL,
        ],

        'modified' => [
          'mysql_type' => 'timestamp',
          'pgsql_type' => 'timestamp',
          'sqlite_type' => 'datetime',
          'not null' => FALSE,
          'default' => NULL,
        ],
      ],

      // Primary Key.
      'primary key' => [
        'entity_id',
      ],

    ];

    return $schema;
  }

  /*--------------------------------------------------------------------
   *
   * The foldershare field operations.
   *
   *-------------------------------------------------------------------*/
  /**
   * Insert the content type about the foldershare into the table.
   *
   * @param array $entity
   *  the multiple entity objects
   *
   * @return void
   * @throws \Exception
   */
  public static function insertContentType(array $entity) {

    // Get the database connection
    $connection = Database::getConnection();

    // who's created the content type for the foldershare
    $userId = \Drupal::currentUser()->id();

    // Add a new entry.
    $query = $connection->insert(self::CONTENT_TYPE_TABLE);
    $query->fields(
      [
        'user_id' => $userId,
        'bundle' => $entity['bundle'],
        'entity_type' => $entity['entity_type'],
        'entity_id' => $entity['entity_id'],
        'fs_id' => $entity['fs_id'],
        'fs_uuid' => $entity['fs_uuid'],
        'created' => $entity['created'],
        'modified' => $entity['modified'],
      ]);
    $query->execute();
  }

  /**
   * Clears all of content types for the foldershare entity.
   *
   * @return void
   */
  public static function clearContentType() {
    $connection = Database::getConnection();
    $connection->delete(self::CONTENT_TYPE_TABLE)->execute();
  }

  /**
   * Deletes the content types for the foldershare entity.
   *
   * @param $entity_id
   *   (optional, default = -1) With a -1, content types are deleted
   *   by forwarding to clearContentType(). Otherwise, each entity ID
   *   given is deleted from the content type table.
   *
   * @return void
   */
  public static function deleteContentType($entity_id = -1) {
    if ($entity_id === -1 || $entity_id === '-1' || $entity_id === '') {
      self::clearContentType();
      return;
    }
    $connection = Database::getConnection();
    $query = $connection->delete(self::CONTENT_TYPE_TABLE);
    $query->condition('entity_id', $entity_id, '=');
    $query->execute();
  }

  /**
   * Returns unsorted content types of all entities.
   *
   * @return array
   */
  public static function &getAllContentTypes() {
    // Query the usage table for all entries.
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'user_id', 'user_id');
    $select->addField('u', 'bundle', 'bundle');
    $select->addField('u', 'entity_type', 'entity_type');
    $select->addField('u', 'entity_id', 'entity_id');
    $select->addField('u', 'fs_id', 'fs_id');
    $select->addField('u', 'fs_uuid', 'fs_uuid');
    $select->addField('u', 'created', 'created');
    $select->addField('u', 'modified', 'modified');
    $select->leftJoin('users_field_data', 'r', 'r.uid = u.user_id');
    $select->addField('r', 'name', 'name');
    $records = $select->execute()->fetchAll();

    // Build and return an array from the records.  Array keys
    // are user IDs, while values are usage info.
    $contentTypes = [];
    foreach ($records as &$record) {
      $contentTypes[(int) $record->entity_id] = [
        'user_id' => &$record->user_id,
        'user_name' => &$record->name,
        'bundle' => &$record->bundle,
        'entity_type' => &$record->entity_type,
        'fs_id' => &$record->fs_id,
        'fs_uuid' => &$record->fs_uuid,
        'created' => &$record->created
      ];
    }

    return $contentTypes;
  }

  /**
   * Check ans see if the content type has enabled the foldershare field
   *
   * @param $entity_id
   *  The entity ID.
   *
   * @return bool
   */
  public static function isContentType($entity_id): bool {
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'entity_id', 'entity_id');
    $select->condition('u.entity_id', $entity_id, '=');
    $select->range(0, 1);
    $records = $select->execute();

    $entityExists = ($records->rowCount() > 0);

    if ($entityExists) {
      return TRUE;
    }
    else {
      return FALSE;
    }
  }

  /**
   * Get the foldershare field in the content type.
   *
   * @param $entity_id
   * The entity ID.
   *
   * @return array
   */
  public static function getContentType($entity_id) {
    // Query the usage table for all entries.
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'user_id', 'user_id');
    $select->addField('u', 'bundle', 'bundle');
    $select->addField('u', 'entity_type', 'entity_type');
    $select->addField('u', 'entity_id', 'entity_id');
    $select->addField('u', 'fs_id', 'fs_id');
    $select->addField('u', 'fs_uuid', 'fs_uuid');
    $select->addField('u', 'created', 'created');
    $select->addField('u', 'modified', 'modified');
    $select->leftJoin('users_field_data', 'r', 'r.uid = u.user_id');
    $select->addField('r', 'name', 'name');
    $select->condition('u.entity_id', $entity_id, '=');
    $select->range(0, 1);
    return $select->execute()->fetchAll();
  }

  /**
   * Get the root foldershare ID for the content type
   *
   * @param $entity_id
   *  The entity ID.
   * @param $currentFoldershareId
   *  The current foldershare ID
   * @param $fileScheme
   *  The file storage
   * @param $fileDirectory
   *  The default file directory to store files
   *
   * @return int|mixed|string|null
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function getRootIdForContentType($entity_id, $currentFoldershareId,
                                                 $fileScheme, $fileDirectory) {
    // Query the foldershare ID for the entity
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'fs_id', 'fs_id');
    $select->condition('u.entity_id', $entity_id, '=');
    $select->range(0, 1);
    $result = $select->execute()->fetchAll();

    $rootFoldershareId = -1;
    if (count($result) > 0) {
      $currentItem = FolderShare::load($result[0]->fs_id);
      if ($currentItem !== NULL) {
        // Update the correct file schema and file directory in foldershare entity
        // for the content type.
        if ($currentItem->getFileScheme() != $fileScheme ||
          $currentItem->getRootBaseFilePath() != $fileDirectory) {
          $currentItem->setFileScheme($fileScheme);
          $currentItem->setRootBaseFilePath($fileDirectory);
          $currentItem->save();
        }
        // Find the root foldershare ID in foldershare entities for the content type.
        $currentItemId = $currentItem->id();
        $descendantIds = $currentItem->findDescendantIds();
        if ($currentItemId === $currentFoldershareId ||
          in_array($currentFoldershareId, $descendantIds)) {
          $rootFoldershareId = $currentItemId;
        }
      }
    }
    return $rootFoldershareId;
  }

  /**
   * Create the FolderShare Field for the entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *  The entity to enable the foldershare field.
   * @param $fileDirectory
   *  The file directory to stores files and folders.
   * @param $allowRename
   *  The rename is allowed
   * @param $ownerUid
   *  The owner of the user.
   * @param $fileScheme
   *  The file storage
   *
   * @return int[]
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function createFolderShareField(EntityInterface $entity,
                                                                $fileDirectory,
                                                                $allowRename,
                                                                $ownerUid,
                                                                $fileScheme = 'public') {
    $entityId = $entity->id();
    $currentFoldershareId = -1;
    $rootFoldershareId = -1;

    // Retrieve the foldershare field for this entity
    $fieldFoldershare = self::getContentType($entityId);
    // If the current user is anonymous user, and there no descendants,
    // the foldershare in the entity is not shown.
    // Otherwise, the foldershare field is displayed in the entity.
    if (\Drupal::currentUser()->isAnonymous()) {
      if (count($fieldFoldershare) != 0) {
        $currentFoldershareId = $fieldFoldershare[0]->fs_id;
        $fieldFolder = FolderShare::load($currentFoldershareId);
        if (count($fieldFolder->findDescendantIds()) == 0) {
          $currentFoldershareId = -1;
        }
        else {
          $rootFoldershareId = $currentFoldershareId;
        }
      }
    }
    else {
      if (count($fieldFoldershare) == 0) {
        // Create the foldershare entity and file directory for the content entity
        $currentFoldershareId = self::createDataDirectory($entity, $fileDirectory,
          $allowRename, $ownerUid, $fileScheme);
        $rootFoldershareId = $currentFoldershareId;
      }
      else {
        // Even if the foldershare ID for this entity is found in the table, if there is no found in
        // the foldershare entities,
        // create the foldershare entity and file directory for the content entity
        $currentFoldershareId = $fieldFoldershare[0]->fs_id;
        $fieldFolder = FolderShare::load($currentFoldershareId);

        if (empty($fieldFolder)) {
          $currentFoldershareId = self::createDataDirectory($entity, $fileDirectory,
            $allowRename, $ownerUid, $fileScheme, TRUE);
          $rootFoldershareId = $currentFoldershareId;
        }
        else {
          // Basically the owner can manage the foldershare browser.
          // But, if users has the 'content_editor' and the 'administrator',
          // they can manage it as well.
          // If there is no descendants, other users can't see the foldershare browser, either.
          $rootFoldershareId = $currentFoldershareId;
          $originalUserName = $fieldFoldershare[0]->name;
          $currentUserName = \Drupal::currentUser()->getAccountName();
          if ($originalUserName !== $currentUserName) {
            $roles = \Drupal::entityTypeManager()
              ->getStorage('user_role')
              ->loadMultiple();
            $currentUserRoles = \Drupal::currentUser()->getRoles();
            // Check user has specific role; example 'administrator' role
            $role_flag = FALSE;
            $allowed_roles = ['content_editor', 'administrator'];
            foreach ($allowed_roles as $role) {
              if (in_array($role, $currentUserRoles)) {
                $role_flag = TRUE;
              }
            }
            if ($role_flag === FALSE) {
              if (count($fieldFolder->findDescendantIds()) == 0) {
                $currentFoldershareId = -1;
              }
            }
          }
        }
      }
    }

    return [$currentFoldershareId, $rootFoldershareId];
  }

  /**
   * Get the root foldershare ID for the entity
   *
   * @param \Drupal\foldershare\FolderShareInterface $foldershare
   *  The foldershare entity
   *
   * @return int|mixed
   */
  public static function getRootFolderId(FolderShareInterface $foldershare) {
    $ancestorFolderIds = $foldershare->findAncestorFolderIds();
    $folderIDs = self::getAllFolderShareIds();
    $rootFoldershareId = array_values(array_intersect($ancestorFolderIds, $folderIDs));
    if (!empty($rootFoldershareId)) {
      return $rootFoldershareId[0];
    }
    return -1;
  }

  /**
   * Retunrs all of foldershare IDs in the content types table.
   *
   * @return array
   */
  public static function getAllFolderShareIds() {
    $foldershareIds = [];
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'fs_id', 'fs_id');
    $result = $select->execute()->fetchAll();
    if (count($result) > 0) {
      foreach ($result as $item) {
        $foldershareIds[] = $item->fs_id;
      }
    }

    return $foldershareIds;
  }

  /**
   * Create the foldershare entity for the content entity,
   * the file storage with the file directory, and the entry for the content type
   * in the content types table.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *  The entity object for the foldershare entity
   * @param $fileDirectory
   *  The file directory to store files and folders
   * @param $allowRename
   *  if it is true, the rename for files and folders is allowed
   * @param $ownerUid
   *  The owner ID.
   * @param $fileScheme
   *  The file storage
   * @param $delete
   *  The delete flag for deleting the existing content type.
   *
   * @return int|mixed|string|null
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\TypedData\Exception\ReadOnlyException
   */
  public static function createDataDirectory(EntityInterface $entity,
                                                             $fileDirectory,
                                                             $allowRename,
                                                             $ownerUid,
                                                             $fileScheme = 'public',
                                                             $delete = FALSE) {
    // Create the foldershare entity and file directory
    $entityId = $entity->id();
    $entityType = $entity->getEntityType()->id();
    $folderName = $entityType . '_' . $entityId;
    $fieldFolder = FolderShare::createRootFolder($folderName, $allowRename, $ownerUid, $fileScheme, $fileDirectory);;
    $fieldFolder->setEnabledContentType(TRUE);
    $fieldFolder->save();

    // Create the content type for the foldershare entity in the content types table.
    $currentFoldershareId = $fieldFolder->id();
    $timestamp = new DrupalDateTime();
    $fieldData = [
      'bundle' => $entity->bundle(),
      'entity_type' => $entityType,
      'entity_id' => $entityId,
      'fs_id' => $currentFoldershareId,
      'fs_uuid' => $fieldFolder->uuid(),
      'created' => $timestamp->format('Y-m-d H:i:s'),
      'modified' => $timestamp->format('Y-m-d H:i:s'),
    ];
    if ($delete) {
      self::deleteContentType($entityId);
    }
    self::insertContentType($fieldData);

    return $currentFoldershareId;
  }

  /**
   * Get the configuration for field and storage settings
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *  The entity object
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getFolderShareConfig(EntityInterface $entity) {
    $config = [];
    $entityId = $entity->id();
    $entityTypeId = $entity->getEntityType()->id();
    $bundle = $entity->bundle();
    // Retrieve the field definitions for the content entity
    $fieldDefinitions = $entity->getFieldDefinitions();
    foreach ($fieldDefinitions as $fieldDefinition) {
      // Find the field for the foldershare entity in the field storage's configuration
      $fieldStorage = $fieldDefinition->getFieldStorageDefinition();
      $fieldName = $fieldStorage->getName();

      // Get the field values for a specific field.
      $fieldValues = self::getFieldSettings($entityTypeId, $bundle, $fieldName);

      /*
       * field key = foldershare_max_filesize
       * field key = foldershare_file_extensions
       * field key = foldershare_folder
       * field key = foldershare_target_type
       * field key = foldershare_uri_scheme
       */
      if (!empty($fieldValues) && isset($fieldValues['foldershare-field-filefolder']['foldershare-field-folder'])) {
        $fileDirectory = self::getFileDirectory($bundle, $entityTypeId, $entityId);
        if ($fileDirectory === NULL) {
          // Get the field value for the file directory.
          $fileDirectory = $fieldValues['foldershare-field-filefolder']['foldershare-field-folder'];
          // Assuming you have the $node object representing your node.
          $node = \Drupal::entityTypeManager()
            ->getStorage($entityTypeId)
            ->load($entityId);

          // Prepare a data array for token replacement.
          $data = [
            $entityTypeId => $node,
          ];

          $fileDirectory = trim($fileDirectory, '/');

          // Replace tokens. As the tokens might contain HTML we convert it to plain
          // text.
          $fileDirectory = PlainTextOutput::renderFromHtml(\Drupal::token()
            ->replace($fileDirectory, $data));
        }
        $config['file_directory'] = $fileDirectory;

        // Get the field value for the file storage
        if (isset($fieldValues['foldershare-field-directories']['section']['foldershare-field-uri-scheme'])) {
          $config['file_uri_scheme'] = $fieldValues['foldershare-field-directories']['section']['foldershare-field-uri-scheme'];
        }

        // Get the field value for the file extensions
        if (isset($fieldValues['foldershare-field-fileextensions']['section']['set-extensions']['foldershare-field-file-extensions'])) {
          $config['file_extensions'] = $fieldValues['foldershare-field-fileextensions']['section']['set-extensions']['foldershare-field-file-extensions'];
        }

        // Get the field value for the file size
        if (isset($fieldValues['foldershare-field-fileuploadsize']['section']['set-file-upload-size']['foldershare-field-file-upload-size'])) {
          $config['file_max_filesize'] = $fieldValues['foldershare-field-fileuploadsize']['section']['set-file-upload-size']['foldershare-field-file-upload-size'];
        }

        // Get the field value for the flag for all file extensions
        if (isset($fieldValues['foldershare-field-fileextensions']['section']['all-file-extensions'])) {
          $config['all_file_extensions'] = $fieldValues['foldershare-field-fileextensions']['section']['all-file-extensions'];
        }

        // Get the field value for the flag for any file upload size
        if (isset($fieldValues['foldershare-field-fileuploadsize']['section']['any-file-upload-size'])) {
          $config['any_file_upload_size'] = $fieldValues['foldershare-field-fileuploadsize']['section']['any-file-upload-size'];
        }
        break;
      }
    }
    return $config;
  }

  /**
   * Get the folsersgare ID as the parameter from the request URL
   *
   * @return bool|float|int|mixed|string
   */
  public static function getCurrentFolderShareID() {
    $currentFoldershareId = -1;
    // Get the foldershare ID from the URL.
    $param = \Drupal::request()->query->get('foldershare-folder-browser-menu-parentId');
    if (!empty($param)) {
      $currentFoldershareId = $param;
    }

    // Get the current request.
    $request = \Drupal::requestStack()->getCurrentRequest();

    // Get the request values.
    $requestValues = $request->request->all();
    if (isset($requestValues["foldershare-folder-browser-menu-parentId"])) {
      $currentFoldershareId = intval($requestValues["foldershare-folder-browser-menu-parentId"]);
    }

    if (isset($requestValues['destinationid'])) {
      $currentFoldershareId = $requestValues['destinationid'];
    }

    return $currentFoldershareId;
  }

  /**
   * Get the field values for a specific field in the configuration
   *
   * @param string $entityType
   *   The entity type the field belongs to.
   * @param string $bundle
   *   The bundle name the field belongs to.
   * @param string $fieldName
   *   The machine name of the field.
   *
   * @return array
   *   The field settings.
   */
  public static function getFieldSettings(string $entityType, string $bundle, string $fieldName) {
    $settings = [];

    // Load the field definition.
    $fieldDefinition = FieldConfig::loadByName($entityType, $bundle, $fieldName);

    if ($fieldDefinition) {
      // If the field is using a custom field type, retrieve additional settings.
      $fieldType = $fieldDefinition->getType();
      if (!empty($fieldType)) {
        if ($fieldType === 'foldershare_field') {
          $settings = $fieldDefinition->getSettings();
        }
      }
    }

    return $settings;
  }

  /**
   * Check and see if the foldershare entities does exist for the content entity.
   *
   * @param $entity_bundle
   *  The bundle name for the entity.
   * @param $entity_type_id
   *  The entity type ID for the entity.
   *
   * @return bool
   */
  public static function hasContents($entity_bundle, $entity_type_id) {
    // Query the content types table to get the foldershare entities for the entity.
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'fs_id', 'fs_id');
    $select->condition('u.bundle', $entity_bundle, '=');
    $select->condition('u.entity_type', $entity_type_id, '=');
    // $select->range(0, 1);
    $result = $select->execute()->fetchAll();

    $hasContents = FALSE;
    if (count($result) > 0) {
      foreach ($result as $item) {
        $currentItem = FolderShare::load($item->fs_id);
        if ($currentItem !== NULL) {
          $childrenIds = $currentItem->findChildrenIds();
          if (count($childrenIds) > 0) {
            $hasContents = TRUE;
            break;
          }
        }
      }
    }

    return $hasContents;
  }

  /**
   * Get the settings configuration for the foldershare entity
   *
   * @param $foldershareId
   *  The foldershare ID to retrieve the content type in the table.
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getFieldConfiguration($foldershareId) {
    // Query the content types table for an entry for this foldershare ID.
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'entity_id', 'entity_id');
    $select->addField('u', 'entity_type', 'entity_type');
    $select->condition('u.fs_id', $foldershareId, '=');
    $select->range(0, 1);
    $result = $select->execute()->fetchAll();

    $config = [];
    if (count($result) > 0) {
      // Load the content type entity from the entity ID
      $entityTypeManager = \Drupal::service('entity_type.manager');
      $entity = $entityTypeManager->getStorage($result[0]->entity_type)
        ->load($result[0]->entity_id);
      if ($entity) {
        // Get the configuration for the field settings and field storage settings for this entity.
        $config = self::getFolderShareConfig($entity);
      }
    }

    return $config;
  }

  /**
   * Create or get the view display name for the content type
   *
   * @param $entity_bundle
   *  The bundle name for the entity. It is used for the view id and the view title name.
   * @param $existing_view_name
   *  The view name to use the foldershare view list, or other view name
   *
   * @return int|mixed|string
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function getViewsPage($entity_bundle, $existing_view_name = 'foldershare_lists') {
    // Validate the specified view name
    $view_list = Views::getAllViews();
    $is_view = FALSE;
    foreach ($view_list as $view) {
      if ($view->id() === $existing_view_name) {
        $is_view = TRUE;
        break;
      }
    }
    if ($is_view) {
      // Get the view display list in the specified view name
      $existing_view = Views::getView($existing_view_name);

      // The "storage" object has the view display list
      $storage = $existing_view->storage;

      // Create new view display ID.
      $new_display_id = $entity_bundle . '_foldershare_field';

      // Check the available view display ID
      $displays = $storage->get('display');

      if (!array_key_exists($new_display_id, $displays)) {
        // Use the existing display name, "list_folder" in the foldershare view list.
        $existing_display_name = 'list_folder';
        // Define the display type, page, block, embed, etc.
        $display_type = 'embed';
        // Create the new view display
        $duplicate_display_id = $storage->duplicateDisplayAsType($existing_display_name, $display_type);
        // Get new view display
        $new_display = &$storage->getDisplay($duplicate_display_id);

        // Change the display id and the display title.
        $new_display['id'] = $new_display_id;
        $new_display['display_title'] = ucfirst($entity_bundle) . ' FolderShare Field';
        $new_display['display_options']['title'] = $new_display['display_title'];
        // Delete the duplicated view display
        unset($displays[$duplicate_display_id]);
        // Add new display
        $displays[$new_display_id] = $new_display;
        $storage->set('display', $displays);
        @$storage->save();
      }
      return $new_display_id;
    }
    else {
      \Drupal::messenger()->addStatus(
        t('View %view_name is not found.', [
          '%view_name' => $existing_view_name
        ]));
      return '';
    }
  }

  /**
   * Get the file extensions and file upload size for the foldershare entity in the content type.
   *
   * @param \Drupal\foldershare\FolderShareInterface|NULL $foldershare
   *  The foldershare entity checks the field settings for the content type.
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getFileSettingsForContentType(FolderShareInterface $foldershare = NULL) {
    $extensions = [];
    $fileUploadSizeLimit = 0;

    if (!empty($foldershare)) {
      // Check the foldershare entity for the content type.
      $isEnabledContentType = $foldershare->getRootItem()->isEnabledContentType();
      if ($isEnabledContentType) {
        // Get the field setting for the foldershare entity
        $rootId = $foldershare->getRootItem()->id();
        $field_config = self::getFieldConfiguration($rootId);

        // If the flag for the all file extensions is true,
        // the default for file extensions is used.
        if ($field_config['all_file_extensions']) {
          $extensionsString = Settings::getAllowedFilenameExtensionsDefault();
        }
        else {
          $extensionsString = $field_config['file_extensions'];
        }
        $extensions = mb_split(' ', $extensionsString);

        // If the flag for the any file upload size is true,
        // the default value for the file upload size is used.
        if ($field_config['any_file_upload_size']) {
          $fileUploadSizeLimit = LimitUtilities::getPhpFileUploadSizeLimit();
        }
        else {
          if ($field_config['file_max_filesize'] <= 0) {
            $fileUploadSizeLimit = LimitUtilities::getPhpFileUploadSizeLimit();
          }
          else {
            $fileUploadSizeLimit = $field_config['file_max_filesize'];
          }
        }
      }
    }
    return [$extensions, $fileUploadSizeLimit];
  }

  public static function getFileDirectory($entity_bundle, $entity_type_id, $entity_id) {
    // Query the content types table to get the foldershare entities for the entity.
    $connection = Database::getConnection();
    $select = $connection->select(self::CONTENT_TYPE_TABLE, 'u');
    $select->addField('u', 'fs_id', 'fs_id');
    $select->condition('u.bundle', $entity_bundle, '=');
    $select->condition('u.entity_type', $entity_type_id, '=');
    $select->condition('u.entity_id', $entity_id, '=');
    $select->range(0, 1);
    $result = $select->execute()->fetchAll();

    $fileDirectory = NULL;
    if (count($result) > 0) {
      $currentItem = FolderShare::load($result[0]->fs_id);
      if ($currentItem !== NULL) {
        $childrenIds = $currentItem->findChildrenIds();
        if (count($childrenIds) > 0) {
          $fileDirectory = $currentItem->getRootBaseFilePath();
        }
      }
    }

    return $fileDirectory;
  }

}
