<?php

namespace Drupal\foldershare;

use Drupal\file\FileInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\foldershare\Entity\FolderShare;
use Drupal\foldershare\Utilities\FileUtilities;

/**
 * Manages file storage in the local file system.
 *
 * @ingroup foldershare
 */
final class ManageFileSystem {

  /*--------------------------------------------------------------------
   *
   * Constants.
   *
   *--------------------------------------------------------------------*/

  /**
   * The name of the top-level folder into which the module places files.
   *
   * The folder is within the public or private file system, depending
   * upon which file system is selected by the site administrator using
   * the module's settings.
   *
   * @var string
   * @see ::getFileUri()
   */
  const FILE_DIRECTORY = 'foldersharefiles';

  /*---------------------------------------------------------------------
   *
   * Temporary files.
   *
   *---------------------------------------------------------------------*/
  /**
   * Creates an empty temporary file in the module's temporary directory.
   *
   * A new empty file is created in the module's temporary files directory.
   * The file has a randomly generated name that does not collide with
   * anything else in the directory. Permissions are set appropriately for
   * web server access.
   *
   * @return string
   *   Returns a URI for a local temp file on success, and NULL on error.
   *
   * @see ::getTempDirectoryUri()
   * @see ::createLocalTempDirectory()
   */
  public static function createLocalTempFile() {
    // Create a temporary file with a random name in the temp directory.
    // The random name starts with the given prefix. Some OSes (e.g. Windows)
    // limit this prefix to three characters.
    $uri = FileUtilities::tempnam(self::getTempDirectoryUri(), 'tmp');
    if ($uri === FALSE) {
      return NULL;
    }

    // Insure the file is accessible.
    if (FileUtilities::chmod($uri) === FALSE) {
      return NULL;
    }

    return $uri;
  }

  /**
   * Creates an empty temporary directory in the module's data directory tree.
   *
   * A new empty directory is created in the module's temporary files directory.
   * The directory has a randomly generated name that does not collide with
   * anything else in the directory. Permissions are set appropriate for
   * web server access.
   *
   * @return string
   *   Returns a URI for a local temp directory on success, and NULL on error.
   *
   * @see ::getTempDirectoryUri()
   * @see ::createLocalTempFile()
   */
  public static function createLocalTempDirectory() {
    // There is no "tempnam()" for directories, only files. Use it to get
    // a safe random name, then replace the file with a directory.
    $uri = self::createLocalTempFile();
    if ($uri === NULL ||
        FileUtilities::unlink($uri) === FALSE ||
        FileUtilities::mkdir($uri) === FALSE ||
        FileUtilities::chmod($uri) === FALSE) {
      return NULL;
    }

    return $uri;
  }

  /**
   * Returns the URI for a temporary directory.
   *
   * @return string
   *   The URI for the directory path.
   */
  public static function getTempDirectoryUri() {
    // Use the Drupal standard "temporary" URI scheme, which goes to the
    // site's configured temp directory.
    return 'temporary://';
  }

  /*---------------------------------------------------------------------
   *
   * URIs.
   *
   *---------------------------------------------------------------------*/
  /**
   * Returns the URI for a file managed by the module.
   *
   * The returned URI has the form "SCHEME://SUBDIR/DIRS.../FILE.EXTENSION"
   * where:
   *
   * - SCHEME is either "public" or "private", based upon the module's file
   *   storage configuration.
   *
   * - SUBDIR is the name of the module's subdirectory for its files within
   *   the public or private file system.
   *
   * - DIRS.../FILE is a chain of numeric subdirectory names and a numeric
   *   file name (see below).
   *
   * - EXTENSION is the filename extension from the user-visible file name.
   *
   * Numeric subdirectory names are based upon the File object's entity ID,
   * and not the user-chosen file name or folder names. Such names have
   * multiple possible problems:
   *
   * - User file and directory names are not guaranteed to be unique site-wide,
   *   so they cannot safely coexist with names from other users unless some
   *   safe organization mechanism is introduced.
   *
   * - User file and directory names may use the full UTF-8 character set
   *   (except for a few restricted characters, like '/', ':', and '\'), but
   *   the underlying file system may not support the same characters.
   *
   * - USer file and diretory names may be up to the maximum length supported
   *   by the module (such as 255 characters), but the underlying file system
   *   may be more restrictive. For instance, file paths (the chain of
   *   directory names from '/' down to and including the file name) may be
   *   limited as well (such as to 1024 characters), which imposes another
   *   limit on directory depth.
   *
   * - User files may be nested within folders, and those folders within
   *   folders, to an arbitrary depth, but the underlying file system may
   *   have depth limits.
   *
   * - Any number of user files may be stored within the same folder, but
   *   the underlying file system may have limits or performance issues with
   *   very large directories.
   *
   * For these reasons, user-chosen file and directory names cannot be used
   * as-is. They must be replaced with alternative names that are safe for
   * any file system and unique.
   *
   * This module creates directory and file names based upon the unique
   * integer entity ID of the File object referencing the file. These IDs
   * are stored in Drupal as 64-bit integers, which can be represented as
   * 20 zero-padded digits (0..9).  These digits are then divided into groups
   * of consecutive digits based upon a module-wide configuration constant
   * DIGITS_PER_SERVER_DIRECTORY_NAME.  A typical value for this constant
   * is 4. This causes a 20-digit number to be split into 5 groups of 4 digits
   * each. The last of these groups is used as a numeric name for the file,
   * while the preceding groups are used as numeric names for subdirectories.
   *
   * As more File objects are created for use by this module, more numeric
   * files are added to the directories. When an entity ID rolls over to
   * update the 5th digit, the file is put into the next numeric subdirectory,
   * and so on. This keeps directory sizes under control (with a maximum of
   * 10^4 = 10,000 files each).
   *
   * The numeric file name has the original file's filename extension appended
   * so that file field formatters and other Drupal tools that look at paths
   * and filenames will see a proper extension.
   *
   * Once the file path is built for the given File entity, any directories
   * needed are created, permissions are set, and .htaccess files are added.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file current stored locally, or soon to be stored locally.
   * @param \Drupal\foldershare\FolderShareInterface|NULL $item
   *   The foldershare entity has the root file path for storing files.
   *
   * @return string
   *   The URI for the directory path to the file.
   *
   * @see ::getTempDirectoryUri()
   * @see ::prepareDirectories()
   * @see \Drupal\foldershare\Settings::getFileScheme()
   */
  /**
   * @param \Drupal\file\FileInterface $file

   *
   * @return string
   */
  public static function getFileUri(FileInterface $file,
                                    FolderShareInterface $item = NULL): string {
    //
    // Use multi-byte character string functions throughout. Local file
    // systems and URIs use UTF-8.
    //
    // Create file path
    // ----------------
    // The file URI is composed of:
    // - The configured public or private scheme.
    // - The configured subdirectory for files within the site.
    // - A chain of numeric directory names for the File entity ID.
    // - A final numeric file name.
    // - The file name extension of the original file.
    //
    // Start the file path.
    $path = "";

    if ($item != NULL) {
      // Get the root file path and file scheme for the foldershare entity
      $fileScheme = $item->getRootItem()->getFileScheme();
      $localFilePath = $item->getRootItem()->getRootBaseFilePath();

      try {
        $path .= $localFilePath . $item->getPath();
        if ($item->isFolder()) {
          $path .= '/' . $file->getFilename();
        }
      } catch (\Exception $e) {
        // Deletion throws an exception if the directory, or any of its
        // descendants, cannot be deleted. This could be due to a permissions
        // problem, an unmounted file system, or perhaps the directory has
        // already been deleted.
        //
        // In any case, there's nothing we can do.
        ManageLog::error(
          "Local file system: Uninstallation of the @module module cannot delete the module's local directory.\nThe directory is at \"@uri\". Attempting to delete it threw exception with the message \"@message\".",
          [
            '@module' => Constants::MODULE,
            '@uri' => $path,
            '@message' => $e->getMessage(),
            'exception' => $e,
          ]);
      }
    } else {
      // Get the default path for the foldershare entity
      $fileScheme = Settings::getFileScheme();
      $homeFolderPath = self::FILE_DIRECTORY . '/' . Settings::getHomeFolderName() . '/' .
        \Drupal::currentUser()->getAccountName();
      $path .= $homeFolderPath . '/' . $file->getFilename();
    }

    //
    // Create directories
    // ------------------
    // Create any specified directories that don't already exist, set their
    // permissions, and add .htaccess files as needed.
    self::prepareFileDirectories($path, $fileScheme);

    // Add "public" or "private" to create the URI.
    return  $fileScheme . '://' . $path;
  }

  /**
   * Delete files or folders on the file system
   *
   * @param FolderShareInterface $item
   * @return void
   */
  public static function delete(FolderShareInterface $item): void
  {
    // Get the root file path and the file scheme for the foldershare entity.
    if ($item->isRootItem()) {
      $homeFolderPath = $item->getRootBaseFilePath();
      $fileScheme = $item->getFileScheme();
    } else {
      $homeFolderPath = $item->getRootItem()->getRootBaseFilePath();
      $fileScheme = $item->getRootItem()->getFileScheme();
    }

    // Set the URI for files and folders,
    // And then remove them.
    $uri = $fileScheme . '://' . $homeFolderPath;
    try {
      $uri .= $item->getPath();
      if ($fileScheme === 's3') {
        $s3Key = $homeFolderPath . $item->getPath();
        ManageS3FileSystem::deleteS3Directory($s3Key);
      } else {
        FileUtilities::rrmdir($uri);
      }
    }
    catch (\Exception $e) {
      // Deletion throws an exception if the directory, or any of its
      // descendants, cannot be deleted. This could be due to a permissions
      // problem, an unmounted file system, or perhaps the directory has
      // already been deleted.
      //
      // In any case, there's nothing we can do.
      ManageLog::error(
        "Local file system: Uninstallation of the @module module cannot delete the module's local directory.\nThe directory is at \"@uri\". Attempting to delete it threw exception with the message \"@message\".",
        [
          '@module'   => Constants::MODULE,
          '@uri'      => $uri,
          '@message'  => $e->getMessage(),
          'exception' => $e,
        ]);
    }
  }

  /**
   * Rename files and folders on the file system
   *
   * @param $targetName
   *  The new name.
   * @param FolderShareInterface|NULL $source
   *  The source foldershare entity. It has the root file path and file scheme
   * for renaming the new name.
   *
   * @return void
   */
  public static function rename($targetName, FolderShareInterface $source): void
  {
    // find the existing uri.
    $old_uri = '';
    try {
      // Get the root file path and file scheme for the source entity.
      if ($source->isRootItem()) {
        $homeFolderPath = $source->getRootBaseFilePath();
        $fileScheme = $source->getFileScheme();
      } else {
        $homeFolderPath = $source->getRootItem()->getRootBaseFilePath();
        $fileScheme = $source->getRootItem()->getFileScheme();
      }
      // Get the file storage and the actual file location
      $basicUri = $fileScheme . '://';
      $folder_path = $homeFolderPath;
      if ($source->getParentFolder() != null) {
        $folder_path .= $source->getParentFolder()->getPath();
      }

      // Set up the source and destination uris
      $old_uri = $basicUri . $folder_path . '/' . $source->getName();
      $new_uri = $basicUri . $folder_path . '/' . $targetName;
      if ($fileScheme !== 's3') {
        FileUtilities::rename($old_uri, $new_uri);
      } else {
        ManageS3FileSystem::s3Move($old_uri, $new_uri, $source->isFolder());
      }
    }
    catch (\Exception $e) {
      // Deletion throws an exception if the directory, or any of its
      // descendants, cannot be deleted. This could be due to a permissions
      // problem, an unmounted file system, or perhaps the directory has
      // already been deleted.
      //
      // In any case, there's nothing we can do.
      ManageLog::error(
        "Local file system: Uninstallation of the @module module cannot delete the module's local directory.\nThe directory is at \"@uri\". Attempting to delete it threw exception with the message \"@message\".",
        [
          '@module'   => Constants::MODULE,
          '@uri'      => $old_uri,
          '@message'  => $e->getMessage(),
          'exception' => $e,
        ]);
    }
  }

  /**
   * Move files and folders on the file system.
   *
   * @param \Drupal\foldershare\FolderShareInterface $source
   *  The source foldershare entity
   * @param \Drupal\foldershare\FolderShareInterface $destination
   *  The destination foldershare entity
   *
   * @return void
   */
  public static function move(FolderShareInterface $source, FolderShareInterface $destination): void
  {

    try {
      // Get the root file path and the file scheme for source's foldershare entity
      if ($source->isRootItem()) {
        $sourceHomeFolderPath = $source->getRootBaseFilePath();
        $sourceFileScheme = $source->getFileScheme();
      } else {
        $sourceHomeFolderPath = $source->getRootItem()->getRootBaseFilePath();
        $sourceFileScheme = $source->getRootItem()->getFileScheme();
      }

      // Get the root file path and the file scheme for destination's foldershare entity
      if ($destination->isRootItem()) {
        $destHomeFolderPath = $destination->getRootBaseFilePath();
        $destFileScheme = $destination->getFileScheme();
      } else {
        $destHomeFolderPath = $destination->getRootItem()->getRootBaseFilePath();
        $destFileScheme = $destination->getRootItem()->getFileScheme();
      }

      // Set up the URI for source and destination
      $src_uri = $sourceFileScheme . '://' . $sourceHomeFolderPath . $source->getPath();
      $dest_uri = $destFileScheme . '://' . $destHomeFolderPath . $destination->getPath();

      // Rename source's URI to destination's URI
      if ($sourceFileScheme !== 's3') {
        FileUtilities::rename($src_uri, $dest_uri);
      } else {
        ManageS3FileSystem::s3Move($src_uri, $dest_uri, $source->isFolder());
      }
    }
    catch (\Exception $e) {
      // Deletion throws an exception if the directory, or any of its
      // descendants, cannot be deleted. This could be due to a permissions
      // problem, an unmounted file system, or perhaps the directory has
      // already been deleted.
      //
      // In any case, there's nothing we can do.
      ManageLog::error(
        "Local file system: Uninstallation of the @module module cannot delete the module's local directory.\nThe directory is at \"@uri\". Attempting to delete it threw exception with the message \"@message\".",
        [
          '@module'   => Constants::MODULE,
          '@uri'      => $source->getPath(),
          '@message'  => $e->getMessage(),
          'exception' => $e,
        ]);
    }
  }

  /**
   * Find the file and the folder on the file system
   *
   * @param $path
   *  The path for the file or the folder
   * @param string|NULL $fileScheme
   *  The file storage
   * @param string|NULL $homeFolderPath
   *  The actual file location for file ot folder
   *
   * @return bool
   */
  public static function fileExists($path='',
                                    string $fileScheme = NULL,
                                    string $homeFolderPath = NULL): bool
  {
    $manager = \Drupal::service('stream_wrapper_manager');
    if (is_null($fileScheme)) {
      $fileScheme = Settings::getFileScheme();
    }
    $stream = $manager->getViaScheme($fileScheme);

    if ($stream === FALSE) {
      // Fail with unknown scheme.
      return FALSE;
    }
    $isLocalStream = FileUtilities::isLocalStreamWrapper($stream->getType());

    try {
      $streamDirectory = $stream->getDirectoryPath();
    } catch (\Throwable $e) { // For PHP 7
      $streamDirectory = '';
    } catch (\Exception $e) { // For PHP 5
      $streamDirectory = '';
    }

    // Set the default actual file location if the home folder path is not set.
    if (is_null($homeFolderPath)) {
      $homeFolderPath = ManageFileSystem::FILE_DIRECTORY . '/' . Settings::getHomeFolderName();
    }
    $realPath = $streamDirectory . '/' . $homeFolderPath;

    if ($path !== '') {
      if (str_starts_with($path, '/')) {
        $realPath .= $path;
      }
      else {
        $realPath .= '/' . $path;
      }
    }

    // Check and see if the file exists.
    if ($isLocalStream) {
      if (FileUtilities::fileExists($realPath) === FALSE) {
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * Update the file uri or all file uris in the folder in the database
   *
   * @param \Drupal\foldershare\FolderShareInterface $item
   *  The foldershare entity for the file ro folder
   *
   * @return void
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function updateFileUri(FolderShareInterface $item): void
  {
    if ($item->isFolder()) {
      $descendant_ids = $item->findDescendantIds();
      foreach ($descendant_ids as $descendant_id) {
        $descendant = FolderShare::load($descendant_id);

        // For a file or image, update the corresponding File entity's name
        // and then, if needed, update the name of the stored file.
        if (!$descendant->isFolder()) {
          if (!$descendant->isMedia()) {
            self::setFileUri($descendant);
          }
        }
      }
    } else {
      self::setFileUri($item);
    }
  }

  /**
   * Set the file uri and save it in the database
   *
   * @param FolderShareInterface $item
   *  The foldershare entity
   *
   * @return void
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private static function setFileUri(FolderShareInterface $item): void
  {
    if ($item->isFile() === TRUE) {
      $file = $item->getFile();
    } else {
      $file = $item->getImage();
    }

    // Get the root file path and the file scheme in the foldershaare entity
    if ($item->isRootItem()) {
      $homeFolderPath = $item->getRootBaseFilePath();
      $fileScheme = $item->getFileScheme();
    } else {
      $homeFolderPath = $item->getRootItem()->getRootBaseFilePath();
      $fileScheme = $item->getRootItem()->getFileScheme();
    }

    $file->setFileUri($fileScheme . '://' . $homeFolderPath . $item->getPath());
    $file->save();
  }

  /**
   *
   *
   * @param string $targetName
   * @return bool
   */
  /**
   * Create the folder on the file system
   *
   * @param string $targetName
   *  The targeted folder name
   * @param string|NULL $fileScheme
   *  The file storage
   * @param string|NULL $homeFolderPath
   *  The actual file location for the folder
   *
   * @return bool
   */
  public static function createFolder(string $targetName,
                                      string $fileScheme = NULL,
                                      string $homeFolderPath = NULL): bool
  {

    // If the file scheme is not set up, the default file scheme is configured.
    if (is_null($fileScheme)) {
      $fileScheme = Settings::getFileScheme();
    }

    // If the actual file location is not set up, the default file location is configured.
    if (is_null($homeFolderPath)) {
      $homeFolderPath = self::FILE_DIRECTORY . '/' . Settings::getHomeFolderName();
    }
    $basicPath = $homeFolderPath . '/' . $targetName;

    //
    // Create directories
    // ------------------
    // Create any specified directories that don't already exist, set their
    // permissions, and add .htaccess files as needed.
    return self::prepareFileDirectories($basicPath, $fileScheme, false);
  }

  /**
   * Creates module file directories and adds .htaccess files.
   *
   * The given URI is converted to a file system path and all missing
   * directories on the path created and their permissions set.
   *
   * If the module's file directory does not have a .htaccess file, or
   * there is a file but it may have been edited, then a new .htaccess
   * file is added that block's direct web server access.
   *
   * @param string $path
   *   The local file path for a file under management by this module.
   * @param string $fileScheme
   *   The file storage.
   * @param $parent
   *   If the flag is true, the parent's path is set up,
   *   if the flag is false, the current path is set up.
   *
   * @return bool
   *   Returns TRUE on success and FALSE on an error. The only error case
   *   occurs when a directory cannot be created, and a message is logged.
   *
   * @see \Drupal\Core\File\FileSystemInterface::prepareDirectory()
   */
  private static function prepareFileDirectories(string $path, string $fileScheme, $parent=true) {
    //
    // Setup.
    // ------
    // Get the current public/private file system choice for file storage,
    // then build a path to the parent directory.
    $manager = \Drupal::service('stream_wrapper_manager');
    $fileSystem = \Drupal::service('file_system');

    $stream = $manager->getViaScheme($fileScheme);

    if ($stream === FALSE) {
      // Fail with unknown scheme.
      return FALSE;
    }

    $isLocalStream = FileUtilities::isLocalStreamWrapper($stream->getType());

    try {
        $streamDirectory = $stream->getDirectoryPath();
    } catch (\Throwable $e) { // For PHP 7
        $streamDirectory = '';
    } catch (\Exception $e) { // For PHP 5
        $streamDirectory = '';
    }

    // If the parent flag is true, the parent's path is set up,
    // if the parent flag is false, the current path is set up.
    if ($parent) {
      $parentPath = FileUtilities::dirname($streamDirectory . '/' . $path);

      // Compute the parent path URI, which is basically the dirname of the $path prefixed with: "<stream wrapper's scheme>://"
      $parentPathURI = FileUtilities::dirname("$fileScheme://$path");
    } else {
      $parentPath = $streamDirectory . '/' . $path;
      $parentPathURI = "$fileScheme://$path";
    }

    //
    // Create parent directories, if needed.
    // -------------------------------------
    // If one or more parent directories on the file's path do not exist yet.
    // Create them, setting permissions.
    if (!$isLocalStream || FileUtilities::fileExists($parentPath) === FALSE) {
      if ($fileSystem->prepareDirectory(
          $parentPathURI,
        (FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) === FALSE) {
        // The parent directory creation failed. There are several possible
        // reasons for this:
        // - The directory already exists.
        // - The file system is full.
        // - The file system is offline.
        // - Something horrific is going on.
        //
        // Of these, only the first case is interesting. While above we
        // checked if the directory already exists, it is possible that
        // another process has snuck in and created the directory before
        // we've tried to create it. That will cause a failure here.
        //
        // Ideally, that other process was an instance of Drupal using
        // FolderShare. In that case the file prepare would have occurred
        // correctly, with permissions set, etc. But we can't be sure.
        // So, let's try again to be sure, but only if the directory now
        // appears to exist.
        $failure = TRUE;
        if (!$isLocalStream || FileUtilities::fileExists($parentPath) === TRUE) {
          if ($fileSystem->prepareDirectory(
                    $parentPathURI,
                    (FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) === TRUE) {
            // It worked this time!
            $failure = FALSE;
          }
        }

        if ($failure === TRUE) {
          // The parent directories could not be created.
          ManageLog::critical(
            "Local file system: '@path' could not be created.\nThe directory is needed by the @module module to store uploaded files. There may be a problem with the server's file system directories and/or permissions.",
            [
              '@path'   => $parentPath,
              '@module' => Constants::MODULE,
            ]);
          return FALSE;
        }
      }
      elseif ($isLocalStream && FileUtilities::fileExists($parentPath) === FALSE) {
        // The parent directory creation succeeded, but the directory isn't
        // there? This should not be possible.
        ManageLog::critical(
          "Local file system: '@path' could not be read.\nThe directory is needed by the @module module to store uploaded files. There may be a problem with the server's file system permissions.",
          [
            '@path'   => $parentPath,
            '@module' => Constants::MODULE,
          ]);
        return FALSE;
      }
    }

    if (!in_array($fileScheme, ['private', 's3'])) {
      //
      // Create .htaccess file in top directory, if needed.
      // --------------------------------------------------
      // Drupal can be configured to use one or both of:
      // - a public (default) directory served directly by the web server.
      // - a private directory out of view from the web server.
      //
      // For a private directory, Drupal requires and adds a .htaccess file
      // that insures that if (somehow) a web server can see the directory,
      // everything within the directory is blocked from being directly served
      // by the web server. Instead, file accesses are redirected to Drupal,
      // which can do access control.
      //
      // For a public directory, Drupal adds a .htaccess file that allows
      // a web server to directly serve the files. File accesses then do not
      // invoke Drupal, and there is no access control.
      //
      // FolderShare always wants access control. When the module is configured
      // to use a private directory, there is nothing more required. Drupal's
      // default .htaccess file already directs accesses back to Drupal, even
      // for subdirectories and files deep below.
      //
      // When FolderShare is configured to use a public directory, however,
      // the Drupal default .htaccess at the top of the files directory tree
      // is not suitable. The module's subdirectory needs its own .htaccess
      // file that closes down direct web server access. Thereafter, accesses
      // to files anywhwere in the module's files subdirectory cause redirects
      // to Drupal where we can do access control.
      //
      // SECURITY NOTE:
      // - A .htaccess file is not required when using a private file system.
      //   We assume that the private file system is, indeed, private and that
      //   a web server therefore cannot see the files stored there, or an
      //   .htaccess file stored there.
      //
      // - On the first use of the directory, there will be no .htaccess file
      //   and one must be created.
      //
      // - On further use of the directory, it is possible that a hacker or
      //   misguided site admin has edited the .htaccess file to open access.
      //   This is not appropriate since the files in this directory are
      //   strictly under FolderShare management. To prevent a possible
      //   access control bypass, the .htaccess file is OVERWRITTEN every
      //   time this function is called.
      $filesPath = $streamDirectory . '/' . self::FILE_DIRECTORY;

      // file_save_access's third argument (TRUE) indicates the function
      // should overwrite any existing .htaccess file. Unfortunately, it
      // does not actually do this... if the file doesn't have write
      // permission. In order to force the issue, we need to first change
      // the file's permission to gain write permission, if the file exists.
      $htaccess = $filesPath . '/.htaccess';
      if ($isLocalStream && FileUtilities::fileExists($htaccess) === TRUE) {
        @chmod($htaccess, 0666);
      }

      // DEPRECATED. Remove this when Drupal 8.7 is no longer supported.
      $versionParts = explode('.', \Drupal::VERSION);
      $drupalVersion = floatval($versionParts[0] . '.' . $versionParts[1]);
      if ($drupalVersion < 8.8) {
        // Pre-Drupal 8.8, the htaccess file writer service did not exist.
        // Call the deprecated file_save_htaccess() function.
        call_user_func('file_save_htaccess', $filesPath, TRUE, TRUE);
      }
      // END DEPRECATED.
      else {
        $writer = \Drupal::service('file.htaccess_writer');
        $writer->write($filesPath, TRUE, TRUE);
      }

      // SECURITY NOTE:
      // - It is possible that a hacker or site admin has added a .htaccess
      //   file to intermediate directories. These should be removed, but
      //   detecting them requires a top-down directory tree traversal,
      //   which is expensive.
    }

    // The parent directories now all exist, though the file does not yet.
    return TRUE;
  }

  /**
   * Parses a URI for a module file and returns the File entity ID.
   *
   * @param string $path
   *   The local file path for a file managed by this module.
   *
   * @return int|bool
   *   Returns the integer File entity ID parsed from the path, or FALSE if
   *   the path is not for a file managed by this module.
   */
  public static function getFileEntityId(string $path) {
    //
    // Find start of numeric directory names
    // -------------------------------------
    // Look for use of the module's FILE_DIRECTORY name. If not found,
    // the path is malformed.
    $fileDirectory = self::FILE_DIRECTORY;
    $pos = mb_strpos($path, $fileDirectory);
    if ($pos === FALSE) {
      // Fail. FILE_DIRECTORY is not in the path, therefore this is not a
      // path to a module file.
      return FALSE;
    }

    //
    // Parse entity ID
    // ---------------
    // Strip the file directory from the front of the path, then use the
    // remainder to build the 20 digit entity ID. Return the integer form
    // of that ID.
    $remainder = mb_substr($path, ($pos + mb_strlen($fileDirectory) + 1));

    // Remove all of the slashes, joining together the numeric names of
    // the subdirectories.
    $digits = mb_ereg_replace('/', '', $remainder);

    // Remove the file extension, if any.
    $dotIndex = mb_strrpos($digits, '.');
    if ($dotIndex !== FALSE) {
      $digits = mb_substr($digits, 0, $dotIndex);
    }

    if (mb_strlen($digits) !== 20) {
      // Fail. Wrong number of digits. Not directory names and file name.
      return FALSE;
    }

    // Convert the numeric string to a file entity ID.
    if (is_numeric($digits) === FALSE) {
      // Fail. Malformed directory and file names.
      return FALSE;
    }

    return (int) intval($digits);
  }

}
