<?php

namespace Drupal\foldershare;

use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Symfony\Component\HttpFoundation\StreamedResponse;

/**
 * Manages file storage in the S3 file system.
 *
 * @ingroup foldershare
 */
final class ManageS3FileSystem {

  /**
   * Move or rename a directory in S3.
   *
   * @param string $sourceUri
   *   The URI of the source directory, e.g., 's3://old-directory/'.
   * @param string $destinationUri
   *   The URI of the destination directory, e.g., 's3://new-directory/'.
   * @param bool $isFolder
   *   (optional, default = FALSE) When TRUE, the URI of the source and destination is the directory.
   *
   * @throws \Exception
   *   Throws exception if the move operation fails.
   */
  public static function s3Move(string $sourceUri, string $destinationUri, bool $isFolder = FALSE): bool {
    $fileSystem = \Drupal::service('file_system');

    // Check if source exists.
    if (!ManageS3FileSystem::s3FileExists($sourceUri)) {
      \Drupal::logger("foldershare s3")->debug("The source path, $sourceUri, does not exist.");
      return FALSE;
    }

    if ($isFolder) {
      // Ensure source and destination URIs end with a slash for directories.
      $sourceUri = rtrim($sourceUri, '/') . '/';
      $destinationUri = rtrim($destinationUri, '/') . '/';

      // Move all files in the source directory.
      $files = $fileSystem->scanDirectory($sourceUri, '/.*/', ['recurse' => TRUE]);

      if (!empty($files)) {
        foreach ($files as $file) {
          $sourcePath = $file->uri;
          $relativePath = str_replace($sourceUri, '', $sourcePath);
          $destinationPath = $destinationUri . $relativePath;

          if (!self::moveFile($sourcePath, $destinationPath)) {
            // Stop on first failure.
            return FALSE;
          }
        }
        return self::deleteS3Directory($sourceUri);
      }
    } else {
      // Move single file.
      return self::moveFile($sourceUri, $destinationUri);
    }
    return TRUE;
  }

  /**
   * Helper function to move a file from source to destination.
   *
   * @param string $sourcePath
   *   The URI of the source file.
   * @param string $destinationPath
   *   The URI of the destination file.
   *
   * @return bool
   *   TRUE on success, FALSE on failure.
   * @throws \Exception
   */
  private static function moveFile(string $sourcePath, string $destinationPath): bool {
    if (self::copyFileInS3($sourcePath, $destinationPath)) {
      // Delete the original file.
      if (\Drupal::service('file_system')->delete($sourcePath)) {
        return TRUE;
      } else {
        \Drupal::logger("foldershare s3")->debug("Failed to delete original file: $sourcePath");
        return FALSE;
      }
    } else {
      \Drupal::logger("foldershare s3")->debug("Failed to copy file from $sourcePath to $destinationPath");
      return FALSE;
    }
  }

  /**
   * Copies a file within an S3 bucket.
   *
   * @param string $sourceUri
   *   The source URI of the file, e.g.,
   *   's3://foldersharefiles/users/cyoun/samples/test.jpg'.
   * @param string $destinationUri
   *   The destination URI, e.g.,
   *   's3://foldersharefiles/users/cyoun/examples/test.jpg'.
   *
   * @throws \Exception
   *   Throws exception if the copy operation fails.
   */
  public static function copyFileInS3(string $sourceUri, string $destinationUri): bool {
    $file_system = \Drupal::service('file_system');

    // Extract the directory path from the source and destination URIs.
    $sourceDir = dirname($sourceUri);
    $destinationDir = dirname($destinationUri);

    // Check if the source directory is accessible.
    if (!$file_system->prepareDirectory($sourceDir, FileSystemInterface::CREATE_DIRECTORY)) {
      // throw new \Exception("The source file '$sourceUri' does not exist or cannot be accessed.");
      \Drupal::logger("foldershare s3")->debug("The source file '$sourceUri' does not exist or cannot be accessed.");
      return FALSE;
    }

    // Check if the destination directory is accessible or needs to be created.
    if (!$file_system->prepareDirectory($destinationDir, FileSystemInterface::CREATE_DIRECTORY)) {
      // throw new \Exception("The destination directory '$destinationUri' cannot be created or accessed.");
      \Drupal::logger("foldershare s3")->debug("The destination directory '$destinationUri' cannot be created or accessed.");
      return FALSE;
    }

    // Copy the file to the new location.
    if (!$file_system->copy($sourceUri, $destinationUri, FileExists::Replace)) {
      // throw new \Exception("Failed to copy file from '$sourceUri' to '$destinationUri'.");
      \Drupal::logger("foldershare s3")->debug("Failed to copy file from '$sourceUri' to '$destinationUri'.");
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Check if the path is available in S3 bucket.
   *
   * @param string $path
   *   The URI of the file or the directory
   *
   * @return bool
   */
  public static function s3FileExists(string $path): bool {
    // Get the stream wrapper manager service.
    $stream_wrapper_manager = \Drupal::service('stream_wrapper_manager');

    // Check if the path is a valid URI.
    if (!$stream_wrapper_manager->isValidUri($path)) {
      return FALSE;
    }

    // Get the file system service.
    $file_system = \Drupal::service('file_system');

    // Check if the path is a directory.
    if ($file_system->prepareDirectory($path, FileSystemInterface::MODIFY_PERMISSIONS)) {
      // Path is a valid directory.
      return TRUE;
    }

    // Attempt to open the file using file_get_contents().
    try {
      $contents = @file_get_contents($path);
      return $contents !== FALSE;
    }
    catch (\Exception $e) {
      // If there's an error (e.g., file not found), return FALSE.
      \Drupal::logger('foldershare s3')->error("Exception during file exists operation: {$e->getMessage()}");
      return FALSE;
    }
  }

  /**
   * Delete directory prefixes in S3 storage
   *
   * @param $directoryPrefix
   *   root directory prefix
   *
   * @return bool
   * @throws \Drupal\s3fs\S3fsException
   */
  public static function deleteS3Directory($directoryPrefix): bool {
    /** @var \Drupal\Core\Config\ConfigFactoryInterface $configFactory */
    $configFactory = \Drupal::service('config.factory');
    /** @var \Drupal\s3fs\S3fsServiceInterface $s3fsService */
    $s3fsService = \Drupal::service('s3fs');
    // Retrieve S3 configuration from Drupal's config system.
    $config = $configFactory->get('s3fs.settings')->get();

    // Initialize the S3 client using the s3fs service.
    $s3_client = $s3fsService->getAmazonS3Client($config);

    // Remove 's3://' from the directory prefix, if present.
    if (strpos($directoryPrefix, 's3://') === 0) {
      // Strip the 's3://' prefix to get the valid path inside the S3 bucket.
      $directoryPrefix = substr($directoryPrefix, 5);
    }

    // Ensure the directory prefix ends with a '/'
    $directoryPrefix = rtrim($directoryPrefix, '/') . '/';

    try {
      // Start the deletion process
      self::deleteObjects($s3_client, $config['bucket'], $directoryPrefix);

      // Check and delete any remaining empty directory markers
      self::deleteEmptyDirectoryMarkers($s3_client, $config['bucket'], $directoryPrefix);

    } catch (\Exception $e) {
      \Drupal::logger('foldershare s3')->error("Exception during delete operation: {$e->getMessage()}");
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Deletes all objects under the given prefix.
   *
   * @param $s3_client
   *   S3 client interface
   * @param $bucket
   *   S3 buccket
   * @param $prefix
   *   the directory prefix
   *
   * @return void
   */
  private static function deleteObjects($s3_client, $bucket, $prefix): void {
    $iterator_args = [
      'Bucket' => $bucket,
      'Prefix' => $prefix,
    ];

    do {
      $objects = $s3_client->listObjectsV2($iterator_args);

      // Check if objects were returned
      if (!empty($objects['Contents'])) {
        $keysToDelete = [];

        // Collect all object keys for deletion (files).
        foreach ($objects['Contents'] as $object) {
          $keysToDelete[] = ['Key' => $object['Key']];
        }

        // Delete the collected objects.
        if (!empty($keysToDelete)) {
          $s3_client->deleteObjects([
            'Bucket' => $bucket,
            'Delete' => [
              'Objects' => $keysToDelete,
              'Quiet' => true,
            ],
          ]);
        }
      }

      // Continue to list until all objects have been processed.
    } while (!empty($objects['IsTruncated']));
  }

  /**
   * Deletes any remaining empty directory markers.
   *
   * @param $s3_client
   *   S3 client interface
   * @param $bucket
   *   S3 buccket
   * @param $prefix
   *   the directory prefix
   *
   * @return void
   */
  private static function deleteEmptyDirectoryMarkers($s3_client, $bucket, $prefix): void {
    // Prepare arguments for the S3 listObjects call to find directory markers
    $iterator_args = [
      'Bucket' => $bucket,
      // Ensure it ends with a /
      'Prefix' => rtrim($prefix, '/') . '/',
      'Delimiter' => '/',
    ];

    // Retrieve subdirectories
    $objects = $s3_client->listObjectsV2($iterator_args);

    // Check if directory markers were found
    if (!empty($objects['CommonPrefixes'])) {
      foreach ($objects['CommonPrefixes'] as $commonPrefix) {
        // The common prefix represents a directory marker
        $subDirectoryMarker = $commonPrefix['Prefix'];
        // \Drupal::logger('foldershare s3')->notice("Deleting empty directory marker: {$subDirectoryMarker}");

        // Remove the empty directory marker
        $s3_client->deleteObjects([
          'Bucket' => $bucket,
          'Delete' => [
            'Objects' => [['Key' => $subDirectoryMarker]],
            'Quiet' => true,
          ],
        ]);

        // Log the removal of directory markers
        // \Drupal::logger('foldershare s3')->notice("Deleted empty directory marker: {$subDirectoryMarker}");

        // Recursively delete any subdirectories under this marker
        self::deleteEmptyDirectoryMarkers($s3_client, $bucket, $subDirectoryMarker);
      }
    }

    // Finally, attempt to delete the original prefix marker if it remains
    if (!empty($prefix)) {
      $s3_client->deleteObjects([
        'Bucket' => $bucket,
        'Delete' => [
          'Objects' => [['Key' => $prefix]],
          'Quiet' => true,
        ],
      ]);
    }
  }

  /**
   * Retrieves a file from an S3 bucket and returns either its file path or
   * file metadata.
   *
   * This function fetches a file from Amazon S3 using the configured S3 bucket
   * settings. It saves the file content to a temporary file and returns its
   * path. Additionally, if the
   * `$header` flag is set to true, the function will also return the file's
   * metadata, including content type and file size.
   *
   * @param string $fileKey
   *   The key (path) of the file within the S3 bucket. It may start with
   *   's3://', which will be stripped.
   * @param bool $fileData
   *   If true, the function will return an array containing the file body,
   *   content type, and content length. If false, it will return only the file
   *   body.
   *
   * @return string|array|bool
   *   - If `$header` is true, it returns an array with the file path, content
   *   type, and content length.
   *   - If `$header` is false, it returns the file path as a string.
   *   - Returns FALSE if an exception occurs during the S3 file retrieval
   *   process.
   * @throws \Drupal\s3fs\S3fsException
   */
  public static function getS3FileObject(string $fileKey, bool $fileData = false): bool|array|string {
    /** @var \Drupal\Core\Config\ConfigFactoryInterface $configFactory */
    $configFactory = \Drupal::service('config.factory');
    /** @var \Drupal\s3fs\S3fsServiceInterface $s3fsService */
    $s3fsService = \Drupal::service('s3fs');

    // Retrieve S3 configuration settings from Drupal's configuration system.
    $config = $configFactory->get('s3fs.settings')->get();

    // Initialize the S3 client using the s3fs service.
    $s3_client = $s3fsService->getAmazonS3Client($config);

    // Remove the 's3://' prefix from the file key if it's present.
    if (strpos($fileKey, 's3://') === 0) {
      // Strip the 's3://' prefix to get the valid path inside the S3 bucket.
      $fileKey = substr($fileKey, 5);
    }

    try {
      // Retrieve the file from S3 without saving it directly to a local directory.
      $result = $s3_client->getObject([
        // The S3 bucket name as configured in Drupal.
        'Bucket' => $config['bucket'],
        // The file key (path) within the S3 bucket.
        'Key' => $fileKey,
      ]);

      // If the $fileData flag is true, return an array with the file body, content type, and file size.
      if ($fileData) {
        return [
          // file body.
          $result['Body'],
          // The content type of the file (e.g., 'application/pdf').
          $result['ContentType'],
          // The size of the file in bytes.
          $result['ContentLength'],
        ];
      } else {
        // return only file body.
        return $result['Body'];
      }

    } catch (\Exception $e) {
      // Log any exception that occurs during the file retrieval process.
      \Drupal::logger('foldershare s3')->error("Exception during file retrieval: {$e->getMessage()}");

      // Return FALSE if an error occurs.
      return FALSE;
    }
  }

}
