<?php

namespace Drupal\foldershare\Plugin\Field\FieldType;

use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\StreamWrapper\PrivateStream;
use Drupal\Core\StreamWrapper\StreamWrapperInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\Form\FormStateInterface;
use Drupal\foldershare\ManageContentType;
use Drupal\foldershare\ManageHooks;
use Drupal\foldershare\Settings;
use Drupal\foldershare\Utilities\FileUtilities;
use Drupal\foldershare\Utilities\LimitUtilities;
use Drupal\foldershare\Utilities\LinkUtilities;

/**
 * Plugin implementation of the 'foldershare_field' field type.
 * Provides a field type of the foldershare field, and field and storage settings.
 *
 *
 * @FieldType(
 *   id = "foldershare_field",
 *   label = @Translation("FolderShare Field"),
 *   description = @Translation("This field stores the setting of a foldershare."),
 *   default_widget = "foldershare_widget",
 *   default_formatter = "",
 *   list_class = "\Drupal\foldershare\Plugin\Field\FieldType\FolderShareFieldItemList",
 * )
 */

class FolderShareItem extends FieldItemBase {

  /**
   * As default, the file storage location and file scheme are stored
   * in the field setting page.
   *
   * @return array
   */
  public static function defaultStorageSettings() {

    return [
        'foldershare_target_type' => 'foldershare',
        'foldershare-field-directories' => ['section' =>
          ['foldershare-field-uri-scheme' =>
            \Drupal::config('system.file')->get('default_scheme')
          ],
        ],
        'foldershare-field-filefolder' => ['foldershare-field-folder' =>
          '[node:url:args:first]/[date:custom:Y]-[date:custom:m]/[node:nid]'
        ],
      ] + parent::defaultStorageSettings();
  }

  /**
   * This schema function is defined since it is mandatory.
   * It is dummy function which doesn't use for the foldershare field.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_definition
   *
   * @return array
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition)
  {
    $schema = [];

    $schema['columns'] = [
      'foldershare_folder' => [
        'type' => 'varchar',
        'length' => 1024,
      ],
    ];

    return $schema;
  }

  /**
   * This property definition function is defined since it is mandatory.
   * It is dummy function which doesn't use for the foldershare field.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_definition
   *
   * @return array|\Drupal\Core\TypedData\DataDefinitionInterface[]
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties = [];

    $properties['foldershare_folder'] = DataDefinition::create('string')
      ->setLabel(t('Folder Name'))
      ->setDescription(t('The name of the folder.'));
    return $properties;
  }

  /**
   * Create the form for setting up the file scheme and file directory fields.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param $has_data
   *
   * @return array
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data) {

    $element = [];
    //
    // Find installed modules.
    // -----------------------
    // Tab sections vary their presentation based upon what modules are
    // installed at the site.
    $helpInstalled = TRUE;

    //
    // Set up class names.
    // -------------------
    // Use a set of standard class names for tab sections.
    $sectionWrapperClass     = 'foldershare-settings-section-wrapper';
    $sectionClass            = 'foldershare-settings-section';
    $sectionDescriptionClass = 'foldershare-settings-section-description';
    $itemDefaultClass        = 'foldershare-settings-item-default';
    $itemNoteClass           = 'foldershare-settings-item-note';
    $warningClass            = 'foldershare-warning';

    //
    // File system.
    // ------------
    // If the site currently has stored files, then the storage scheme
    // and subdirectory choice cannot be changed.
    //
    // If the site doesn't have stored files, but the private file system
    // has not been configured, then the storage scheme cannot be changed.
    $inUseWarning          = '';
    $noPrivateWarning      = '';
    $folderInUseWarning    = '';
    $storageSchemeDisabled = FALSE;
    $folderDirectoryDisabled = FALSE;

    $entity = $this->getEntity();
    $entity_id = $entity->id();
    $entity_type_id = $entity->getEntityTypeId();
    $entity_bundle = $entity->bundle();

    if (ManageContentType::hasContents($entity_bundle, $entity_type_id) === TRUE) {
      // Create a link to the Drupal core system page to delete all content
      // for the entity type.
      $inUseWarning = [
        '#type'                  => 'html_tag',
        '#tag'                   => 'p',
        '#value'                 => $this->t('The above settings are disabled because there are currently files managed by this module.'),
        '#attributes'            => [
          'class'                => [
            $warningClass,
            $itemNoteClass,
          ],
        ],
      ];
      $storageSchemeDisabled = TRUE;

      // Create a link to the Drupal core system page to delete all content
      // for the entity type.
      $folderInUseWarning = [
        '#type'                  => 'html_tag',
        '#tag'                   => 'p',
        '#value'                 => $this->t('The below setting for file directory is disabled because there are currently files managed by this module.'),
        '#attributes'            => [
          'class'                => [
            $warningClass,
            $itemNoteClass,
          ],
        ],
      ];
      $folderDirectoryDisabled = TRUE;
    }

    $file_scheme_options = [];
    $unsupported_file_scheme_options = [];

    foreach (array_keys(FileUtilities::getAvailableStreamWrappers(StreamWrapperInterface::ALL)) as $scheme) {
      $streamWrapper = FileUtilities::getStreamWrapper($scheme);
      if (FileUtilities::isStreamWrapperTypeSupported($streamWrapper->getType())) {
        $file_scheme_options[$scheme] = $streamWrapper->getName();
      } else {
        $unsupported_file_scheme_options[$scheme] = $streamWrapper->getName();
      }
    }

    if (empty(PrivateStream::basePath()) === TRUE && !$file_scheme_options) {
      if ($helpInstalled === TRUE) {
        $noPrivateWarning = [
          '#type'                  => 'html_tag',
          '#tag'                   => 'p',
          '#value'                 => $this->t(
            'This setting is limited to the <em>Public</em> directory because a <em>Private</em> directory has not been set up yet for the site (see @fileHelpLink and @fileDocLink, @fileSettingsLink settings, and the "file_private_path" setting in the site\'s "settings.php" file).',
            [
              '@fileSettingsLink'  => LinkUtilities::createRouteLink(
                'system.file_system_settings',
                '',
                $this->t('File system')),
              '@fileHelpLink'      => LinkUtilities::createHelpLink(
                'file',
                $this->t('File help')),
              '@fileDocLink'       => LinkUtilities::createDocLink(
                'file',
                $this->t('File documentation')),
            ]),
          '#attributes'            => [
            'class'                => [
              $warningClass,
              $itemNoteClass,
            ],
          ],
        ];
      }
      else {
        $noPrivateWarning = [
          '#type'                  => 'html_tag',
          '#tag'                   => 'p',
          '#value'                 => $this->t(
            'This setting is limited to the <em>Public</em> directory because a <em>Private</em> directory has not been set up yet for the site (see @fileDocLink, @fileSettingsLink settings, and the "file_private_path" setting in the site\'s "settings.php" file).',
            [
              '@fileSettingsLink'  => LinkUtilities::createRouteLink(
                'system.file_system_settings',
                '',
                $this->t('File system')),
              '@fileDocLink'       => LinkUtilities::createDocLink(
                'file',
                $this->t('File documentation')),
            ]),
          '#attributes'            => [
            'class'                => [
              $warningClass,
              $itemNoteClass,
            ],
          ],
        ];
      }
      $storageSchemeDisabled = TRUE;
    }

    $element['foldershare-field-directories'] = [
      '#type'                    => 'details',
      '#title'                   => $this->t('File storage'),
      '#open'                    => TRUE,
      '#attributes'              => [
        'class'                  => [$sectionWrapperClass],
      ],

      'section'                  => [
        '#type'                  => 'container',
        '#attributes'            => [
          'class'                => [$sectionClass],
        ],

        'section-description'    => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $this->t(
            "Files are stored using the stream wrapper configured below. Access controls are enforced the same for all options."),
          '#attributes'          => [
            'class'              => [$sectionDescriptionClass],
          ],
        ],

        'foldershare-field-uri-scheme'       => [
          '#type'                => 'select',
          '#options'             => $file_scheme_options,
          '#default_value'       => $this->getSetting('foldershare-field-directories')['section']['foldershare-field-uri-scheme'],
          '#disabled'            => $storageSchemeDisabled,
        ],

        // The default setting.
        'directory-choice-description' => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $this->t(
            'Default: @default',
            [
              '@default'         => $file_scheme_options[Settings::getFileSchemeDefault()],
            ]),
          '#attributes'          => [
            'class'              => [
              'description',
              $itemDefaultClass,
              $itemNoteClass,
            ],
          ],
        ],
      ],
    ];

    // Add warning if there are things disabled.
    if (empty($noPrivateWarning) === FALSE) {
      $element['foldershare-field-directories']['section']['warning'] = $noPrivateWarning;
    }
    elseif (empty($inUseWarning) === FALSE) {
      $element['foldershare-field-directories']['section']['warning'] = $inUseWarning;
    }

    if ($unsupported_file_scheme_options) {
      $unsupported_file_scheme_options_text = implode(", ", array_values($unsupported_file_scheme_options));
      $element['foldershare-field-directories']['section']['unsupported_file_scheme_warning'] = [
        '#type'                  => 'html_tag',
        '#tag'                   => 'p',
        '#value'                 => $this->t(
          "<b>Warning:</b><br/>The following stream wrappers were found but are not of a supported type: $unsupported_file_scheme_options_text<br/>Only stream wrappers that can handle reading & writing visible files can be used."
        ),
        '#attributes'            => [
          'class'                => [
            $warningClass,
            $itemNoteClass,
          ],
        ],
      ];
    }

    $raw_file_scheme = Settings::getRawFileScheme();
    if ($raw_file_scheme !== Settings::getFileScheme()) {
      // the originally configured file scheme is no longer available
      $element['foldershare-field-directories']['section']['file_scheme_warning'] = [
        '#type'                  => 'html_tag',
        '#tag'                   => 'p',
        '#value'                 => $this->t(
          "<b>Warning:</b><br/>The originally configured <em>$raw_file_scheme</em> stream wrapper scheme is no longer available. As a result, accessing any file that was using this file scheme will now throw an error.<br/>To fix this, re-install a stream wrapper for the <em>$raw_file_scheme</em> scheme."
        ),
        '#attributes'            => [
          'class'                => [
            $warningClass,
            $itemNoteClass,
          ],
        ],
      ];
    }

    $element['foldershare-field-filefolder'] = [
      '#type'                    => 'details',
      '#title'                   => $this->t('File storage location'),
      '#open'                    => TRUE,
      '#attributes'              => [
        'class'                  => [$sectionWrapperClass],
      ],
      'foldershare-field-folder' => [
        '#type' => 'textfield',
        '#title' => $this->t('File directory'),
        '#default_value' => $this->getSetting('foldershare-field-filefolder')['foldershare-field-folder'],
        '#description' => $this->t('Optional subdirectory within the upload destination where files will be stored.'),
        '#disabled' => $folderDirectoryDisabled,
        '#weight' => 3,
      ],
    ];

    if (empty($folderInUseWarning) === FALSE) {
      $element['foldershare-field-filefolder']['warning'] = $folderInUseWarning;
    }

    $element['foldershare-field-directories']['#attached']['library'][] = 'foldershare/foldershare.admin';
    return $element;
  }

  /**
   * As default, the file extensions and file size are stored
   * in the field edit page.
   *
   * @return array|array[]
   */
  public static function defaultFieldSettings() {
    $fileUploadSizeLimit = Settings::getFileUploadSizeLimitDefault();
    $allowedFilenameExtensions = Settings::getAllowedFilenameExtensionsDefault();
    $fileRestrictExtensions = !Settings::getFileRestrictFilenameExtensionsDefault();
    return [
        'foldershare-field-fileuploadsize' => ['section' =>
          ['set-file-upload-size' =>
            ['foldershare-field-file-upload-size' =>
              $fileUploadSizeLimit
            ],
            'any-file-upload-size' => TRUE,
          ],
        ],
        'foldershare-field-fileextensions' => ['section' =>
            ['set-extensions' =>
              ['foldershare-field-file-extensions' =>
                $allowedFilenameExtensions
              ],
              'all-file-extensions' => $fileRestrictExtensions,
            ]
        ]
      ] + parent::defaultFieldSettings();
  }

  /**
   * Create the form for setting up the file extensions and file size fields.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  public function fieldSettingsForm(array $form, FormStateInterface $form_state) {
    $element = [];
    $settings = $this->getSettings();

    //
    // Set up class names.
    // -------------------
    // Use a set of standard class names for tab sections.
    $sectionWrapperClass     = 'foldershare-settings-section-wrapper';
    $sectionClass            = 'foldershare-settings-section';
    $sectionDescriptionClass = 'foldershare-settings-section-description';
    $itemDefaultClass        = 'foldershare-settings-item-default';
    $itemNoteClass           = 'foldershare-settings-item-note';

    //
    // File upload size
    // ----------------
    // Select whether file uploads should be restricted based on their size,
    // and what maximum size is allowed.
    //
    // Select whether file uploads should be restricted based on their
    // file name extensions, and what extensions are allowed.
    $fileUploadSizeLimit             = $settings['foldershare-field-fileuploadsize']['section']['set-file-upload-size']['foldershare-field-file-upload-size'];
    $phpFileUploadSizeLimit          = LimitUtilities::getPhpFileUploadSizeLimit();
    $formattedPhpFileUploadSizeLimit = number_format($phpFileUploadSizeLimit);
    $allowedFileExtensions = $settings['foldershare-field-fileextensions']['section']['set-extensions']['foldershare-field-file-extensions'];
    $allFileExtensions = $settings['foldershare-field-fileextensions']['section']['all-file-extensions'];
    $anyFileUploadSize = $settings['foldershare-field-fileuploadsize']['section']['any-file-upload-size'];

    // Create default value messages.
    if ($anyFileUploadSize === TRUE || $anyFileUploadSize === 1) {
      $defaultUploadSizeMessage = $this->t(
        "Default: Allow any size up to the PHP maximum of @phplimit bytes.",
        [
          '@phplimit' => $formattedPhpFileUploadSizeLimit,
        ]);
      $fileUploadSizeLimit = Settings::getFileUploadSizeLimitDefault();
    }
    else {
      $defaultUploadSizeMessage = $this->t(
        'Default: @bytes bytes',
        ['@bytes' => $fileUploadSizeLimit]);
    }

    if ($allFileExtensions === FALSE || $allFileExtensions === 0) {
      $defaultExtensionsMessage = $this->t(
        'Default: Restrict files to a list of allowed extensions');
    }
    else {
      $defaultExtensionsMessage = $this->t('Default: Allow all extensions');
      $allowedFileExtensions = Settings::getAllowedFilenameExtensionsDefault();
    }

    // Create hook override messages.
    $hookSizeModules = ManageHooks::getHookFileUploadSizeLimitAlter();
    if (empty($hookSizeModules) === TRUE) {
      $hookSizeModulesMessage = '';
    }
    elseif (count($hookSizeModules) === 1) {
      $hookSizeModulesMessage = $this->t(
        "The %name module has installed an override that can further limit upload sizes for specific folders or users.",
        [
          '%name' => \Drupal::moduleHandler()->getName($hookSizeModules[0]),
        ]);
    }
    else {
      $names = [];
      foreach ($hookSizeModules as $moduleName) {
        $names[] = \Drupal::moduleHandler()->getName($moduleName);
      }
      $hookSizeModulesMessage = $this->t(
        "@num modules have installed overrides that can further limit upload sizes for specific folders or users: %list.",
        [
          '@num'  => count($names),
          '%list' => implode(', ', $names),
        ]);
    }

    $hookExtensionsModules = ManageHooks::getHookAllowedFilenameExtensionsAlter();
    if (empty($hookExtensionsModules) === TRUE) {
      $hookExtensionsModulesMessage = '';
    }
    elseif (count($hookExtensionsModules) === 1) {
      $hookExtensionsModulesMessage = $this->t(
        "The %name module has installed an override that can limit extensions for specific folders or users.",
        [
          '%name' => \Drupal::moduleHandler()->getName($hookExtensionsModules[0]),
        ]);
    }
    else {
      $names = [];
      foreach ($hookExtensionsModules as $moduleName) {
        $names[] = \Drupal::moduleHandler()->getName($moduleName);
      }
      $hookExtensionsModulesMessage = $this->t(
        "@num modules have installed overrides that can limit extensions for specific folders or users: %list.",
        [
          '@num'  => count($names),
          '%list' => implode(', ', $names),
        ]);
    }

    $element['foldershare-field-fileuploadsize'] = [
      '#type'                    => 'details',
      '#title'                   => $this->t('File upload size restrictions'),
      '#open'                    => TRUE,
      '#attributes'              => [
        'class'                  => [$sectionWrapperClass],
      ],
      'section'                  => [
        '#type'                  => 'container',
        '#attributes'            => [
          'class'                => [$sectionClass],
        ],

        'section-description'    => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $this->t(
            "Optionally restrict file uploads to a maximum size, up to the maximum set by the <code>upload_max_filesize</code> PHP directive in the server's <em>php.ini</em> configuration file."),
          '#attributes'          => [
            'class'              => [$sectionDescriptionClass],
          ],
        ],

        // A checkbox to enable/disable limiting file sizes.
        'any-file-upload-size'    => [
          '#type'                => 'checkbox',
          '#title'               => $this->t('Allow any file upload size'),
          '#default_value'       => $anyFileUploadSize, // (intval($fileUploadSizeLimit) === 0),
          '#id'                => 'foldershare-field-any-file-upload-size',
        ],

        // The default setting.
        'any-file-upload-size-default' => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $defaultUploadSizeMessage,
          '#attributes'          => [
            'class'              => [
              'description',
              $itemDefaultClass,
              $itemNoteClass,
            ],
          ],
        ],

        // Specific limit.
        'set-file-upload-size'   => [
          '#type'                => 'fieldset',
          '#title'               => $this->t('Restrict file upload size'),
          '#states'              => [
            'visible'            => [
              'input[id="foldershare-field-any-file-upload-size"]' => [
                'checked'        => FALSE,
              ],
            ],
          ],

          // The upload size limit.
          'foldershare-field-file-upload-size'     => [
            '#type'              => 'number',
            '#title'             => $this->t('Maximum file size:'),
            '#default_value'     => $fileUploadSizeLimit,
            '#required'          => FALSE,
            '#min'               => 0,
            '#max'               => $phpFileUploadSizeLimit,
            '#step'              => 1,
            '#size'              => 24,
            '#description'       => $this->t(
              "Enter a size up to the PHP maximum. A zero uses the maximum. Changing the size does not affect files that are already on the site."),
          ],
        ],

        // A list of modules that override the extension list.
        'file-upload-size-hooks' => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $hookSizeModulesMessage,
          '#attributes'          => [
            'class'              => [
              $itemNoteClass,
            ],
          ],
        ],
      ],
    ];

    if (empty($hookSizeModulesMessage) === TRUE) {
      unset($element['foldershare-field-fileuploadsize']['section']['file-upload-size-hooks']);
    }

    $element['foldershare-field-fileextensions'] = [
      '#type'                    => 'details',
      '#title'                   => $this->t('File upload filename extension restrictions'),
      '#open'                    => TRUE,
      '#attributes'              => [
        'class'                  => [$sectionWrapperClass],
      ],
      'section'                  => [
        '#type'                  => 'container',
        '#attributes'            => [
          'class'                => [$sectionClass],
        ],

        'section-description'    => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $this->t(
            'Optionally restrict file uploads to files with specific extensions.'),
          '#attributes'          => [
            'class'              => [$sectionDescriptionClass],
          ],
        ],

        // A checkbox to enable/disable allowing all extensions.
        'all-file-extensions'    => [
          '#type'                => 'checkbox',
          '#title'               => $this->t('Allow all extensions'),
          '#default_value'       => $allFileExtensions, // ($fileRestrictExtensions === FALSE),
          '#id'                => 'foldershare-field-all-file-extensions',
        ],

        // The default setting.
        'all-file-extensions-default' => [
          '#type'                => 'html_tag',
          '#tag'                 => 'p',
          '#value'               => $defaultExtensionsMessage,
          '#attributes'          => [
            'class'              => [
              'description',
              $itemDefaultClass,
              $itemNoteClass,
            ],
          ],
        ],

        // Specific extensions.
        'set-extensions'         => [
          '#type'                => 'fieldset',
          '#title'               => $this->t('Restrict file uploads extensions'),
          '#states'              => [
            'visible'            => [
              'input[id="foldershare-field-all-file-extensions"]' => [
                'checked'        => FALSE,
              ],
            ],
          ],

          // A list of allowed extensions.
          'foldershare-field-file-extensions'      => [
            '#type'              => 'textarea',
            '#title'             => $this->t('Allowed extensions:'),
            '#default_value'     => $allowedFileExtensions,
            '#required'          => FALSE,
            '#rows'              => 10,
            '#description'       => $this->t(
              'Separate extensions with spaces. Do not include dots. Changing this list does not affect files that are already on the site.'),
            '#attributes'        => [
              // No browsers support this. Disable autocomplete.
              'autocomplete' => 'off',
              // Most browsers support this. Disables automatic capitalization.
              'autocapitalize' => 'none',
              // Most browsers support this. Disables spell check.
              'spellcheck'   => 'false',
              // Safari only supports this. Disables spell check.
              'autocorrect'  => 'off',
              // Firefox for Android only. Names the 'Enter' key on a
              // virtual keybd.
              'mozactionhint' => $this->t('Change'),
            ],
          ],
        ],

        // A list of modules that override the extension list.
        'file-extensions-hooks' => [
          '#type'              => 'html_tag',
          '#tag'               => 'p',
          '#value'             => $hookExtensionsModulesMessage,
          '#attributes'          => [
            'class'              => [
              $itemNoteClass,
            ],
          ],
        ],
      ],
    ];

    if (empty($hookExtensionsModulesMessage) === TRUE) {
      unset($element['foldershare-field-fileextensions']['section']['file-extensions-hooks']);
    }

    $element['foldershare-field-fileextensions']['#attached']['library'][] = 'foldershare/foldershare.admin';
    return $element;
  }
}
