<?php

namespace Drupal\foldershare\Plugin\FolderShareCommand;

use Drupal\Core\Form\FormStateInterface;
use Drupal\filter\Entity\FilterFormat;
use Drupal\foldershare\Entity\Exception\RuntimeExceptionWithMarkup;
use Drupal\foldershare\Entity\FolderShare;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\foldershare\Entity\Exception\ValidationException;

/**
 * Defines a command plugin to edit an entity.
 *
 * The command is a dummy that does not use a configuration and does not
 * execute to change the entity. Instead, this command is used solely to
 * create an entry in command menus and support a redirect to a stand-alone
 * edit form from which the user can alter fields on the entity.
 *
 * Configuration parameters:
 * - 'parentId': the parent folder, if any.
 * - 'selectionIds': selected entities to edit.
 *
 * @ingroup foldershare
 *
 * @FolderShareCommand(
 *  id              = "foldersharecommand_edit",
 *  label           = @Translation("Edit"),
 *  menuNameDefault = @Translation("Edit..."),
 *  menuName        = @Translation("Edit..."),
 *  description     = @Translation("Edit fields of a selected item."),
 *  category        = "edit",
 *  weight          = 0,
 *  parentConstraints = {
 *    "kinds"   = {
 *      "rootlist",
 *      "any",
 *    },
 *    "access"  = "view",
 *  },
 *  selectionConstraints = {
 *    "types"   = {
 *      "parent",
 *      "one",
 *    },
 *    "kinds"   = {
 *      "any",
 *    },
 *    "access"  = "update",
 *  },
 * )
 */
class Edit extends FolderShareCommandBase {

  /*--------------------------------------------------------------------
   *
   * Construction.
   *
   *--------------------------------------------------------------------*/
  /**
   * Creates an instance of the plugin.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The container from which to get services.
   * @param array $configuration
   *   The plugin configuration.
   * @param string $pluginId
   *   The plugin ID.
   * @param mixed $pluginDefinition
   *   The plugin implementation definition.
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $pluginId,
    $pluginDefinition) {
    return new static(
      $container,
      $configuration,
      $pluginId,
      $pluginDefinition);
  }

  /*--------------------------------------------------------------------
   *
   * Validation
   *
   *-------------------------------------------------------------------*/
  /**
   * Validate the parent ID and selection ID to edit the description for an entity.
   *
   * @return void
   */
  public function validateParameters()
  {
    if ($this->parametersValidated === TRUE) {
      // Already validated.
      return;
    }

    //
    // Validate selection.
    // ---------------------
    // There must be a selection ID. It must be a valid ID. It must
    // not be one of the selected items.
    //
    if (empty($this->getSelectionIds())) {
      $selectionId = $this->getParentId();
    } else {
      $selectionId = $this->getSelectionIds()[0];
    }
    // Selection is a specific folder.
    $selection = FolderShare::load($selectionId);

    if ($selection === NULL) {
      // Destination ID is not valid. This should have been caught
      // well before this validation stage.
      // Developer-facing exception message.
      throw new ValidationException(
        __METHOD__ . " was called with an invalid entity ID '$selectionId'.");
    }

    $this->parametersValidated = TRUE;

    // Garbage collect.
    gc_collect_cycles();
  }

  /*--------------------------------------------------------------------
   *
   * Configuration form.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function getDescription(bool $forPage) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getTitle(bool $forPage) {
    $ids = $this->getSelectionIds();
    if (empty($ids) === TRUE) {
      $ids = [$this->getParentId()];
    }

    return $this->t(
      'Edit "@name"',
      [
        '@name' => FolderShare::findNameForId($ids[0]),
      ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getSubmitButtonName() {
    return $this->t('Save');
  }

  /**
   * {@inheritdoc}
   */
  public function hasConfigurationForm() {
    // If there isn't one yet, then a configuration form is required.
   return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $formState) {

    //
    // Disable auto-complete.
    // ----------------------
    // In principal, adding 'autocomplete="off"' to a form (and/or form
    // element) will disable browser autocomplete. In practice, modern
    // browsers ignore it and provide autocomplete on any text field.
    // If the text field is for a user name, then some browsers will invoke
    // password managers (e.g. Apple's keychain) to let you select a user
    // name.
    //
    // Nevertheless, mark the form with our intent that autocomplete NOT
    // be enabled anywhere on the form.
    $form['#attributes']['autocomplete'] = 'off';
    $form['#attributes']['class'][] = 'foldershare-editform';

    // Get the available text formats.
    $textFormats = FilterFormat::loadMultiple();

    // Create an options array for the text formats.
    $formatOptions = [];
    foreach ($textFormats as $textFormat) {
      if ($textFormat->id() === 'plain_text')
        continue;
      $formatOptions[$textFormat->id()] = $textFormat->label();
    }

    // If the destination ID is still NULL, then default to the current parent.
    if (empty($this->getSelectionIds())) {
      $selectionId = $this->getParentId();
      $this->setSelectionIds([$selectionId]);
    } else {
      $selectionId = $this->getSelectionIds()[0];
    }

    $foldershare = FolderShare::load($selectionId);
    $defaultValue = $foldershare->getDescription() ?? '';
    if (empty($foldershare->getDescriptionFormat())) {
      $selectedFormat = 'basic_html';
    } else {
      $selectedFormat = $foldershare->getDescriptionFormat();
    }

    // Add a text format selector and textarea field.
    $form['foldershare_textarea'] = [
      '#title' => 'Description',
      '#type' => 'text_format',
      '#format' => $selectedFormat, // Set the default text format.
      '#format_options' => $formatOptions, // Add the available text formats as options.
      '#description' => 'Enter a description of the item. The description may be empty, brief, or long.',
      '#attributes' => [
        'class' => ['foldershare-textarea'],
        'rows' => 5, // Specify the number of rows as an attribute
      ],

      '#resizable' => TRUE,
      '#default_value' => $defaultValue,
    ];

    $form['#attached']['drupalSettings']['editForm']['parentId'] = $this->getParentId();
    $form['#attached']['drupalSettings']['editForm']['selectionId'] = $selectionId;
    $form['#attached']['drupalSettings']['editForm']['submit'] = TRUE;
    $form['#attached']['drupalSettings']['editForm']['selectionUuid'] = $foldershare->uuid();

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $formState) {

    // Submit button pressed. Validate everything.
    try {
      $this->validateConfiguration();
    }
    catch (RuntimeExceptionWithMarkup $e) {
      $formState->setErrorByName('foldershare_textarea', $e->getMarkup());
    }
    catch (\Exception $e) {
      $formState->setErrorByName('foldershare_textarea', $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $formState) {

    if ($this->isValidated() === TRUE) {
      // Get the submitted value from the textarea field.
      $foldershareText = $formState->getValue(['foldershare_textarea', 'value']);

      // Get the selected text format.
      $textFormat = $formState->getValue(['foldershare_textarea', 'format']);
      $foldershare = FolderShare::load($this->getSelectionIds()[0]);
      $foldershare = $foldershare->setDescription($foldershareText);
      $foldershare = $foldershare->setDescriptionFormat($textFormat);
      $foldershare->save();
    }
  }

  /*---------------------------------------------------------------------
   *
   * Execution behavior.
   *
   *---------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function getExecuteBehavior() {
    // When there is a selection, execution changes that selection on the
    // current page. While columns may change, the page doesn't, so we
    // only need to refresh the view.
    return FolderShareCommandInterface::POST_EXECUTE_PAGE_REFRESH;
  }

  /*--------------------------------------------------------------------
   *
   * Execute.
   *
   *--------------------------------------------------------------------*/
  /**
   * {@inheritdoc}
   */
  public function execute() {
    // Do nothing. This is never called because of the pre-execute redirect.
  }

}
